/*
 * Copyright (c) 2013-14 Mikko Mononen memon@inside.org
 *
 * This software is provided 'as-is', without any express or implied
 * warranty.  In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 * claim that you wrote the original software. If you use this software
 * in a product, an acknowledgment in the product documentation would be
 * appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 * misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 *
 * The SVG parser is based on Anti-Grain Geometry 2.4 SVG example
 * Copyright (C) 2002-2004 Maxim Shemanarev (McSeem) (http://www.antigrain.com/)
 *
 * Arc calculation code based on canvg (https://code.google.com/p/canvg/)
 *
 * Bounding box calculation based on http://blog.hackers-cafe.net/2009/06/how-to-calculate-bezier-curves-bounding.html
 *
 */

#ifndef NANOSVG_H
#define NANOSVG_H

#ifndef NANOSVG_CPLUSPLUS
#ifdef __cplusplus
extern "C" {
#endif
#endif

// NanoSVG is a simple stupid single-header-file SVG parse. The output of the parser is a list of cubic bezier shapes.
//
// The library suits well for anything from rendering scalable icons in your editor application to prototyping a game.
//
// NanoSVG supports a wide range of SVG features, but something may be missing, feel free to create a pull request!
//
// The shapes in the SVG images are transformed by the viewBox and converted to specified units.
// That is, you should get the same looking data as your designed in your favorite app.
//
// NanoSVG can return the paths in few different units. For example if you want to render an image, you may choose
// to get the paths in pixels, or if you are feeding the data into a CNC-cutter, you may want to use millimeters.
//
// The units passed to NanoSVG should be one of: 'px', 'pt', 'pc' 'mm', 'cm', or 'in'.
// DPI (dots-per-inch) controls how the unit conversion is done.
//
// If you don't know or care about the units stuff, "px" and 96 should get you going.


/* Example Usage:
	// Load SVG
	wxm_NSVGimage* image;
	image = wxm_nsvgParseFromFile("test.svg", "px", 96);
	printf("size: %f x %f\n", image->width, image->height);
	// Use...
	for (wxm_NSVGshape *shape = image->shapes; shape != NULL; shape = shape->next) {
		for (wxm_NSVGpath *path = shape->paths; path != NULL; path = path->next) {
			for (int i = 0; i < path->npts-1; i += 3) {
				float* p = &path->pts[i*2];
				drawCubicBez(p[0],p[1], p[2],p[3], p[4],p[5], p[6],p[7]);
			}
		}
	}
	// Delete
	wxm_nsvgDelete(image);
*/

enum wxm_NSVGpaintType {
	wxm_NSVG_PAINT_UNDEF = -1,
	wxm_NSVG_PAINT_NONE = 0,
	wxm_NSVG_PAINT_COLOR = 1,
	wxm_NSVG_PAINT_LINEAR_GRADIENT = 2,
	wxm_NSVG_PAINT_RADIAL_GRADIENT = 3
};

enum wxm_NSVGspreadType {
	wxm_NSVG_SPREAD_PAD = 0,
	wxm_NSVG_SPREAD_REFLECT = 1,
	wxm_NSVG_SPREAD_REPEAT = 2
};

enum wxm_NSVGlineJoin {
	wxm_NSVG_JOIN_MITER = 0,
	wxm_NSVG_JOIN_ROUND = 1,
	wxm_NSVG_JOIN_BEVEL = 2
};

enum wxm_NSVGlineCap {
	wxm_NSVG_CAP_BUTT = 0,
	wxm_NSVG_CAP_ROUND = 1,
	wxm_NSVG_CAP_SQUARE = 2
};

enum wxm_NSVGfillRule {
	wxm_NSVG_FILLRULE_NONZERO = 0,
	wxm_NSVG_FILLRULE_EVENODD = 1
};

enum wxm_NSVGflags {
	wxm_NSVG_FLAGS_VISIBLE = 0x01
};

typedef struct wxm_NSVGgradientStop {
	unsigned int color;
	float offset;
} wxm_NSVGgradientStop;

typedef struct wxm_NSVGgradient {
	float xform[6];
	char spread;
	float fx, fy;
	int nstops;
	wxm_NSVGgradientStop stops[1];
} wxm_NSVGgradient;

typedef struct wxm_NSVGpaint {
	signed char type;
	union {
		unsigned int color;
		wxm_NSVGgradient* gradient;
	};
} wxm_NSVGpaint;

typedef struct wxm_NSVGpath
{
	float* pts;					// Cubic bezier points: x0,y0, [cpx1,cpx1,cpx2,cpy2,x1,y1], ...
	int npts;					// Total number of bezier points.
	char closed;				// Flag indicating if shapes should be treated as closed.
	float bounds[4];			// Tight bounding box of the shape [minx,miny,maxx,maxy].
	struct wxm_NSVGpath* next;		// Pointer to next path, or NULL if last element.
} wxm_NSVGpath;

typedef struct wxm_NSVGshape
{
	char id[64];				// Optional 'id' attr of the shape or its group
	wxm_NSVGpaint fill;				// Fill paint
	wxm_NSVGpaint stroke;			// Stroke paint
	float opacity;				// Opacity of the shape.
	float strokeWidth;			// Stroke width (scaled).
	float strokeDashOffset;		// Stroke dash offset (scaled).
	float strokeDashArray[8];	// Stroke dash array (scaled).
	char strokeDashCount;		// Number of dash values in dash array.
	char strokeLineJoin;		// Stroke join type.
	char strokeLineCap;			// Stroke cap type.
	float miterLimit;			// Miter limit
	char fillRule;				// Fill rule, see wxm_NSVGfillRule.
	unsigned char flags;		// Logical or of wxm_NSVG_FLAGS_* flags
	float bounds[4];			// Tight bounding box of the shape [minx,miny,maxx,maxy].
	char fillGradient[64];		// Optional 'id' of fill gradient
	char strokeGradient[64];	// Optional 'id' of stroke gradient
	float xform[6];				// Root transformation for fill/stroke gradient
	wxm_NSVGpath* paths;			// Linked list of paths in the image.
	struct wxm_NSVGshape* next;		// Pointer to next shape, or NULL if last element.
} wxm_NSVGshape;

typedef struct wxm_NSVGimage
{
	float width;				// Width of the image.
	float height;				// Height of the image.
	wxm_NSVGshape* shapes;			// Linked list of shapes in the image.
} wxm_NSVGimage;

// Parses SVG file from a file, returns SVG image as paths.
wxm_NSVGimage* wxm_nsvgParseFromFile(const char* filename, const char* units, float dpi);

// Parses SVG file from a null terminated string, returns SVG image as paths.
// Important note: changes the string.
wxm_NSVGimage* wxm_nsvgParse(char* input, const char* units, float dpi);

// Duplicates a path.
wxm_NSVGpath* wxm_nsvgDuplicatePath(wxm_NSVGpath* p);

// Deletes an image.
void wxm_nsvgDelete(wxm_NSVGimage* image);

#ifndef NANOSVG_CPLUSPLUS
#ifdef __cplusplus
}
#endif
#endif

#ifdef NANOSVG_IMPLEMENTATION

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#define wxm_NSVG_PI (3.14159265358979323846264338327f)
#define wxm_NSVG_KAPPA90 (0.5522847493f)	// Length proportional to radius of a cubic bezier handle for 90deg arcs.

#define wxm_NSVG_ALIGN_MIN 0
#define wxm_NSVG_ALIGN_MID 1
#define wxm_NSVG_ALIGN_MAX 2
#define wxm_NSVG_ALIGN_NONE 0
#define wxm_NSVG_ALIGN_MEET 1
#define wxm_NSVG_ALIGN_SLICE 2

#define wxm_NSVG_NOTUSED(v) do { (void)(1 ? (void)0 : ( (void)(v) ) ); } while(0)
#define wxm_NSVG_RGB(r, g, b) (((unsigned int)r) | ((unsigned int)g << 8) | ((unsigned int)b << 16))

#ifdef _MSC_VER
	#pragma warning (disable: 4996) // Switch off security warnings
	#pragma warning (disable: 4100) // Switch off unreferenced formal parameter warnings
	#ifdef __cplusplus
	#define wxm_NSVG_INLINE inline
	#else
	#define wxm_NSVG_INLINE
	#endif
#else
	#define wxm_NSVG_INLINE inline
#endif


static int wxm_nsvg__isspace(char c)
{
	return strchr(" \t\n\v\f\r", c) != 0;
}

static int wxm_nsvg__isdigit(char c)
{
	return c >= '0' && c <= '9';
}

static wxm_NSVG_INLINE float wxm_nsvg__minf(float a, float b) { return a < b ? a : b; }
static wxm_NSVG_INLINE float wxm_nsvg__maxf(float a, float b) { return a > b ? a : b; }


// Simple XML parser

#define wxm_NSVG_XML_TAG 1
#define wxm_NSVG_XML_CONTENT 2
#define wxm_NSVG_XML_MAX_ATTRIBS 256

static void wxm_nsvg__parseContent(char* s,
							   void (*contentCb)(void* ud, const char* s),
							   void* ud)
{
	// Trim start white spaces
	while (*s && wxm_nsvg__isspace(*s)) s++;
	if (!*s) return;

	if (contentCb)
		(*contentCb)(ud, s);
}

static void wxm_nsvg__parseElement(char* s,
							   void (*startelCb)(void* ud, const char* el, const char** attr),
							   void (*endelCb)(void* ud, const char* el),
							   void* ud)
{
	const char* attr[wxm_NSVG_XML_MAX_ATTRIBS];
	int nattr = 0;
	char* name;
	int start = 0;
	int end = 0;
	char quote;

	// Skip white space after the '<'
	while (*s && wxm_nsvg__isspace(*s)) s++;

	// Check if the tag is end tag
	if (*s == '/') {
		s++;
		end = 1;
	} else {
		start = 1;
	}

	// Skip comments, data and preprocessor stuff.
	if (!*s || *s == '?' || *s == '!')
		return;

	// Get tag name
	name = s;
	while (*s && !wxm_nsvg__isspace(*s)) s++;
	if (*s) { *s++ = '\0'; }

	// Get attribs
	while (!end && *s && nattr < wxm_NSVG_XML_MAX_ATTRIBS-3) {
		char* name = NULL;
		char* value = NULL;

		// Skip white space before the attrib name
		while (*s && wxm_nsvg__isspace(*s)) s++;
		if (!*s) break;
		if (*s == '/') {
			end = 1;
			break;
		}
		name = s;
		// Find end of the attrib name.
		while (*s && !wxm_nsvg__isspace(*s) && *s != '=') s++;
		if (*s) { *s++ = '\0'; }
		// Skip until the beginning of the value.
		while (*s && *s != '\"' && *s != '\'') s++;
		if (!*s) break;
		quote = *s;
		s++;
		// Store value and find the end of it.
		value = s;
		while (*s && *s != quote) s++;
		if (*s) { *s++ = '\0'; }

		// Store only well formed attributes
		if (name && value) {
			attr[nattr++] = name;
			attr[nattr++] = value;
		}
	}

	// List terminator
	attr[nattr++] = 0;
	attr[nattr++] = 0;

	// Call callbacks.
	if (start && startelCb)
		(*startelCb)(ud, name, attr);
	if (end && endelCb)
		(*endelCb)(ud, name);
}

int wxm_nsvg__parseXML(char* input,
				   void (*startelCb)(void* ud, const char* el, const char** attr),
				   void (*endelCb)(void* ud, const char* el),
				   void (*contentCb)(void* ud, const char* s),
				   void* ud)
{
	char* s = input;
	char* mark = s;
	int state = wxm_NSVG_XML_CONTENT;
	while (*s) {
		if (*s == '<' && state == wxm_NSVG_XML_CONTENT) {
			// Start of a tag
			*s++ = '\0';
			wxm_nsvg__parseContent(mark, contentCb, ud);
			mark = s;
			state = wxm_NSVG_XML_TAG;
		} else if (*s == '>' && state == wxm_NSVG_XML_TAG) {
			// Start of a content or new tag.
			*s++ = '\0';
			wxm_nsvg__parseElement(mark, startelCb, endelCb, ud);
			mark = s;
			state = wxm_NSVG_XML_CONTENT;
		} else {
			s++;
		}
	}

	return 1;
}


/* Simple SVG parser. */

#define wxm_NSVG_MAX_ATTR 128

enum wxm_NSVGgradientUnits {
	wxm_NSVG_USER_SPACE = 0,
	wxm_NSVG_OBJECT_SPACE = 1
};

#define wxm_NSVG_MAX_DASHES 8

enum wxm_NSVGunits {
	wxm_NSVG_UNITS_USER,
	wxm_NSVG_UNITS_PX,
	wxm_NSVG_UNITS_PT,
	wxm_NSVG_UNITS_PC,
	wxm_NSVG_UNITS_MM,
	wxm_NSVG_UNITS_CM,
	wxm_NSVG_UNITS_IN,
	wxm_NSVG_UNITS_PERCENT,
	wxm_NSVG_UNITS_EM,
	wxm_NSVG_UNITS_EX
};

typedef struct wxm_NSVGcoordinate {
	float value;
	int units;
} wxm_NSVGcoordinate;

typedef struct wxm_NSVGlinearData {
	wxm_NSVGcoordinate x1, y1, x2, y2;
} wxm_NSVGlinearData;

typedef struct wxm_NSVGradialData {
	wxm_NSVGcoordinate cx, cy, r, fx, fy;
} wxm_NSVGradialData;

typedef struct wxm_NSVGgradientData
{
	char id[64];
	char ref[64];
	signed char type;
	union {
		wxm_NSVGlinearData linear;
		wxm_NSVGradialData radial;
	};
	char spread;
	char units;
	float xform[6];
	int nstops;
	wxm_NSVGgradientStop* stops;
	struct wxm_NSVGgradientData* next;
} wxm_NSVGgradientData;

typedef struct wxm_NSVGattrib
{
	char id[64];
	float xform[6];
	unsigned int fillColor;
	unsigned int strokeColor;
	float opacity;
	float fillOpacity;
	float strokeOpacity;
	char fillGradient[64];
	char strokeGradient[64];
	float strokeWidth;
	float strokeDashOffset;
	float strokeDashArray[wxm_NSVG_MAX_DASHES];
	int strokeDashCount;
	char strokeLineJoin;
	char strokeLineCap;
	float miterLimit;
	char fillRule;
	float fontSize;
	unsigned int stopColor;
	float stopOpacity;
	float stopOffset;
	char hasFill;
	char hasStroke;
	char visible;
} wxm_NSVGattrib;

typedef struct wxm_NSVGparser
{
	wxm_NSVGattrib attr[wxm_NSVG_MAX_ATTR];
	int attrHead;
	float* pts;
	int npts;
	int cpts;
	wxm_NSVGpath* plist;
	wxm_NSVGimage* image;
	wxm_NSVGgradientData* gradients;
	wxm_NSVGshape* shapesTail;
	float viewMinx, viewMiny, viewWidth, viewHeight;
	int alignX, alignY, alignType;
	float dpi;
	char pathFlag;
	char defsFlag;
} wxm_NSVGparser;

static void wxm_nsvg__xformIdentity(float* t)
{
	t[0] = 1.0f; t[1] = 0.0f;
	t[2] = 0.0f; t[3] = 1.0f;
	t[4] = 0.0f; t[5] = 0.0f;
}

static void wxm_nsvg__xformSetTranslation(float* t, float tx, float ty)
{
	t[0] = 1.0f; t[1] = 0.0f;
	t[2] = 0.0f; t[3] = 1.0f;
	t[4] = tx; t[5] = ty;
}

static void wxm_nsvg__xformSetScale(float* t, float sx, float sy)
{
	t[0] = sx; t[1] = 0.0f;
	t[2] = 0.0f; t[3] = sy;
	t[4] = 0.0f; t[5] = 0.0f;
}

static void wxm_nsvg__xformSetSkewX(float* t, float a)
{
	t[0] = 1.0f; t[1] = 0.0f;
	t[2] = tanf(a); t[3] = 1.0f;
	t[4] = 0.0f; t[5] = 0.0f;
}

static void wxm_nsvg__xformSetSkewY(float* t, float a)
{
	t[0] = 1.0f; t[1] = tanf(a);
	t[2] = 0.0f; t[3] = 1.0f;
	t[4] = 0.0f; t[5] = 0.0f;
}

static void wxm_nsvg__xformSetRotation(float* t, float a)
{
	float cs = cosf(a), sn = sinf(a);
	t[0] = cs; t[1] = sn;
	t[2] = -sn; t[3] = cs;
	t[4] = 0.0f; t[5] = 0.0f;
}

static void wxm_nsvg__xformMultiply(float* t, float* s)
{
	float t0 = t[0] * s[0] + t[1] * s[2];
	float t2 = t[2] * s[0] + t[3] * s[2];
	float t4 = t[4] * s[0] + t[5] * s[2] + s[4];
	t[1] = t[0] * s[1] + t[1] * s[3];
	t[3] = t[2] * s[1] + t[3] * s[3];
	t[5] = t[4] * s[1] + t[5] * s[3] + s[5];
	t[0] = t0;
	t[2] = t2;
	t[4] = t4;
}

static void wxm_nsvg__xformInverse(float* inv, float* t)
{
	double invdet, det = (double)t[0] * t[3] - (double)t[2] * t[1];
	if (det > -1e-6 && det < 1e-6) {
		wxm_nsvg__xformIdentity(t);
		return;
	}
	invdet = 1.0 / det;
	inv[0] = (float)(t[3] * invdet);
	inv[2] = (float)(-t[2] * invdet);
	inv[4] = (float)(((double)t[2] * t[5] - (double)t[3] * t[4]) * invdet);
	inv[1] = (float)(-t[1] * invdet);
	inv[3] = (float)(t[0] * invdet);
	inv[5] = (float)(((double)t[1] * t[4] - (double)t[0] * t[5]) * invdet);
}

static void wxm_nsvg__xformPremultiply(float* t, float* s)
{
	float s2[6];
	memcpy(s2, s, sizeof(float)*6);
	wxm_nsvg__xformMultiply(s2, t);
	memcpy(t, s2, sizeof(float)*6);
}

static void wxm_nsvg__xformPoint(float* dx, float* dy, float x, float y, float* t)
{
	*dx = x*t[0] + y*t[2] + t[4];
	*dy = x*t[1] + y*t[3] + t[5];
}

static void wxm_nsvg__xformVec(float* dx, float* dy, float x, float y, float* t)
{
	*dx = x*t[0] + y*t[2];
	*dy = x*t[1] + y*t[3];
}

#define wxm_NSVG_EPSILON (1e-12)

static int wxm_nsvg__ptInBounds(float* pt, float* bounds)
{
	return pt[0] >= bounds[0] && pt[0] <= bounds[2] && pt[1] >= bounds[1] && pt[1] <= bounds[3];
}


static double wxm_nsvg__evalBezier(double t, double p0, double p1, double p2, double p3)
{
	double it = 1.0-t;
	return it*it*it*p0 + 3.0*it*it*t*p1 + 3.0*it*t*t*p2 + t*t*t*p3;
}

static void wxm_nsvg__curveBounds(float* bounds, float* curve)
{
	int i, j, count;
	double roots[2], a, b, c, b2ac, t, v;
	float* v0 = &curve[0];
	float* v1 = &curve[2];
	float* v2 = &curve[4];
	float* v3 = &curve[6];

	// Start the bounding box by end points
	bounds[0] = wxm_nsvg__minf(v0[0], v3[0]);
	bounds[1] = wxm_nsvg__minf(v0[1], v3[1]);
	bounds[2] = wxm_nsvg__maxf(v0[0], v3[0]);
	bounds[3] = wxm_nsvg__maxf(v0[1], v3[1]);

	// Bezier curve fits inside the convex hull of it's control points.
	// If control points are inside the bounds, we're done.
	if (wxm_nsvg__ptInBounds(v1, bounds) && wxm_nsvg__ptInBounds(v2, bounds))
		return;

	// Add bezier curve inflection points in X and Y.
	for (i = 0; i < 2; i++) {
		a = -3.0 * v0[i] + 9.0 * v1[i] - 9.0 * v2[i] + 3.0 * v3[i];
		b = 6.0 * v0[i] - 12.0 * v1[i] + 6.0 * v2[i];
		c = 3.0 * v1[i] - 3.0 * v0[i];
		count = 0;
		if (fabs(a) < wxm_NSVG_EPSILON) {
			if (fabs(b) > wxm_NSVG_EPSILON) {
				t = -c / b;
				if (t > wxm_NSVG_EPSILON && t < 1.0-wxm_NSVG_EPSILON)
					roots[count++] = t;
			}
		} else {
			b2ac = b*b - 4.0*c*a;
			if (b2ac > wxm_NSVG_EPSILON) {
				t = (-b + sqrt(b2ac)) / (2.0 * a);
				if (t > wxm_NSVG_EPSILON && t < 1.0-wxm_NSVG_EPSILON)
					roots[count++] = t;
				t = (-b - sqrt(b2ac)) / (2.0 * a);
				if (t > wxm_NSVG_EPSILON && t < 1.0-wxm_NSVG_EPSILON)
					roots[count++] = t;
			}
		}
		for (j = 0; j < count; j++) {
			v = wxm_nsvg__evalBezier(roots[j], v0[i], v1[i], v2[i], v3[i]);
			bounds[0+i] = wxm_nsvg__minf(bounds[0+i], (float)v);
			bounds[2+i] = wxm_nsvg__maxf(bounds[2+i], (float)v);
		}
	}
}

static wxm_NSVGparser* wxm_nsvg__createParser(void)
{
	wxm_NSVGparser* p;
	p = (wxm_NSVGparser*)malloc(sizeof(wxm_NSVGparser));
	if (p == NULL) goto error;
	memset(p, 0, sizeof(wxm_NSVGparser));

	p->image = (wxm_NSVGimage*)malloc(sizeof(wxm_NSVGimage));
	if (p->image == NULL) goto error;
	memset(p->image, 0, sizeof(wxm_NSVGimage));

	// Init style
	wxm_nsvg__xformIdentity(p->attr[0].xform);
	memset(p->attr[0].id, 0, sizeof p->attr[0].id);
	p->attr[0].fillColor = wxm_NSVG_RGB(0,0,0);
	p->attr[0].strokeColor = wxm_NSVG_RGB(0,0,0);
	p->attr[0].opacity = 1;
	p->attr[0].fillOpacity = 1;
	p->attr[0].strokeOpacity = 1;
	p->attr[0].stopOpacity = 1;
	p->attr[0].strokeWidth = 1;
	p->attr[0].strokeLineJoin = wxm_NSVG_JOIN_MITER;
	p->attr[0].strokeLineCap = wxm_NSVG_CAP_BUTT;
	p->attr[0].miterLimit = 4;
	p->attr[0].fillRule = wxm_NSVG_FILLRULE_NONZERO;
	p->attr[0].hasFill = 1;
	p->attr[0].visible = 1;

	return p;

error:
	if (p) {
		if (p->image) free(p->image);
		free(p);
	}
	return NULL;
}

static void wxm_nsvg__deletePaths(wxm_NSVGpath* path)
{
	while (path) {
		wxm_NSVGpath *next = path->next;
		if (path->pts != NULL)
			free(path->pts);
		free(path);
		path = next;
	}
}

static void wxm_nsvg__deletePaint(wxm_NSVGpaint* paint)
{
	if (paint->type == wxm_NSVG_PAINT_LINEAR_GRADIENT || paint->type == wxm_NSVG_PAINT_RADIAL_GRADIENT)
		free(paint->gradient);
}

static void wxm_nsvg__deleteGradientData(wxm_NSVGgradientData* grad)
{
	wxm_NSVGgradientData* next;
	while (grad != NULL) {
		next = grad->next;
		free(grad->stops);
		free(grad);
		grad = next;
	}
}

static void wxm_nsvg__deleteParser(wxm_NSVGparser* p)
{
	if (p != NULL) {
		wxm_nsvg__deletePaths(p->plist);
		wxm_nsvg__deleteGradientData(p->gradients);
		wxm_nsvgDelete(p->image);
		free(p->pts);
		free(p);
	}
}

static void wxm_nsvg__resetPath(wxm_NSVGparser* p)
{
	p->npts = 0;
}

static void wxm_nsvg__addPoint(wxm_NSVGparser* p, float x, float y)
{
	if (p->npts+1 > p->cpts) {
		p->cpts = p->cpts ? p->cpts*2 : 8;
		p->pts = (float*)realloc(p->pts, p->cpts*2*sizeof(float));
		if (!p->pts) return;
	}
	p->pts[p->npts*2+0] = x;
	p->pts[p->npts*2+1] = y;
	p->npts++;
}

static void wxm_nsvg__moveTo(wxm_NSVGparser* p, float x, float y)
{
	if (p->npts > 0) {
		p->pts[(p->npts-1)*2+0] = x;
		p->pts[(p->npts-1)*2+1] = y;
	} else {
		wxm_nsvg__addPoint(p, x, y);
	}
}

static void wxm_nsvg__lineTo(wxm_NSVGparser* p, float x, float y)
{
	float px,py, dx,dy;
	if (p->npts > 0) {
		px = p->pts[(p->npts-1)*2+0];
		py = p->pts[(p->npts-1)*2+1];
		dx = x - px;
		dy = y - py;
		wxm_nsvg__addPoint(p, px + dx/3.0f, py + dy/3.0f);
		wxm_nsvg__addPoint(p, x - dx/3.0f, y - dy/3.0f);
		wxm_nsvg__addPoint(p, x, y);
	}
}

static void wxm_nsvg__cubicBezTo(wxm_NSVGparser* p, float cpx1, float cpy1, float cpx2, float cpy2, float x, float y)
{
	if (p->npts > 0) {
		wxm_nsvg__addPoint(p, cpx1, cpy1);
		wxm_nsvg__addPoint(p, cpx2, cpy2);
		wxm_nsvg__addPoint(p, x, y);
	}
}

static wxm_NSVGattrib* wxm_nsvg__getAttr(wxm_NSVGparser* p)
{
	return &p->attr[p->attrHead];
}

static void wxm_nsvg__pushAttr(wxm_NSVGparser* p)
{
	if (p->attrHead < wxm_NSVG_MAX_ATTR-1) {
		p->attrHead++;
		memcpy(&p->attr[p->attrHead], &p->attr[p->attrHead-1], sizeof(wxm_NSVGattrib));
	}
}

static void wxm_nsvg__popAttr(wxm_NSVGparser* p)
{
	if (p->attrHead > 0)
		p->attrHead--;
}

static float wxm_nsvg__actualOrigX(wxm_NSVGparser* p)
{
	return p->viewMinx;
}

static float wxm_nsvg__actualOrigY(wxm_NSVGparser* p)
{
	return p->viewMiny;
}

static float wxm_nsvg__actualWidth(wxm_NSVGparser* p)
{
	return p->viewWidth;
}

static float wxm_nsvg__actualHeight(wxm_NSVGparser* p)
{
	return p->viewHeight;
}

static float wxm_nsvg__actualLength(wxm_NSVGparser* p)
{
	float w = wxm_nsvg__actualWidth(p), h = wxm_nsvg__actualHeight(p);
	return sqrtf(w*w + h*h) / sqrtf(2.0f);
}

static float wxm_nsvg__convertToPixels(wxm_NSVGparser* p, wxm_NSVGcoordinate c, float orig, float length)
{
	wxm_NSVGattrib* attr = wxm_nsvg__getAttr(p);
	switch (c.units) {
		case wxm_NSVG_UNITS_USER:		return c.value;
		case wxm_NSVG_UNITS_PX:			return c.value;
		case wxm_NSVG_UNITS_PT:			return c.value / 72.0f * p->dpi;
		case wxm_NSVG_UNITS_PC:			return c.value / 6.0f * p->dpi;
		case wxm_NSVG_UNITS_MM:			return c.value / 25.4f * p->dpi;
		case wxm_NSVG_UNITS_CM:			return c.value / 2.54f * p->dpi;
		case wxm_NSVG_UNITS_IN:			return c.value * p->dpi;
		case wxm_NSVG_UNITS_EM:			return c.value * attr->fontSize;
		case wxm_NSVG_UNITS_EX:			return c.value * attr->fontSize * 0.52f; // x-height of Helvetica.
		case wxm_NSVG_UNITS_PERCENT:	return orig + c.value / 100.0f * length;
		default:					return c.value;
	}
	return c.value;
}

static wxm_NSVGgradientData* wxm_nsvg__findGradientData(wxm_NSVGparser* p, const char* id)
{
	wxm_NSVGgradientData* grad = p->gradients;
	if (id == NULL || *id == '\0')
		return NULL;
	while (grad != NULL) {
		if (strcmp(grad->id, id) == 0)
			return grad;
		grad = grad->next;
	}
	return NULL;
}

static wxm_NSVGgradient* wxm_nsvg__createGradient(wxm_NSVGparser* p, const char* id, const float* localBounds, float *xform, signed char* paintType)
{
	wxm_NSVGgradientData* data = NULL;
	wxm_NSVGgradientData* ref = NULL;
	wxm_NSVGgradientStop* stops = NULL;
	wxm_NSVGgradient* grad;
	float ox, oy, sw, sh, sl;
	int nstops = 0;
	int refIter;

	data = wxm_nsvg__findGradientData(p, id);
	if (data == NULL) return NULL;

	// TODO: use ref to fill in all unset values too.
	ref = data;
	refIter = 0;
	while (ref != NULL) {
		wxm_NSVGgradientData* nextRef = NULL;
		if (stops == NULL && ref->stops != NULL) {
			stops = ref->stops;
			nstops = ref->nstops;
			break;
		}
		nextRef = wxm_nsvg__findGradientData(p, ref->ref);
		if (nextRef == ref) break; // prevent infite loops on malformed data
		ref = nextRef;
		refIter++;
		if (refIter > 32) break; // prevent infite loops on malformed data
	}
	if (stops == NULL) return NULL;

	grad = (wxm_NSVGgradient*)malloc(sizeof(wxm_NSVGgradient) + sizeof(wxm_NSVGgradientStop)*(nstops-1));
	if (grad == NULL) return NULL;

	// The shape width and height.
	if (data->units == wxm_NSVG_OBJECT_SPACE) {
		ox = localBounds[0];
		oy = localBounds[1];
		sw = localBounds[2] - localBounds[0];
		sh = localBounds[3] - localBounds[1];
	} else {
		ox = wxm_nsvg__actualOrigX(p);
		oy = wxm_nsvg__actualOrigY(p);
		sw = wxm_nsvg__actualWidth(p);
		sh = wxm_nsvg__actualHeight(p);
	}
	sl = sqrtf(sw*sw + sh*sh) / sqrtf(2.0f);

	if (data->type == wxm_NSVG_PAINT_LINEAR_GRADIENT) {
		float x1, y1, x2, y2, dx, dy;
		x1 = wxm_nsvg__convertToPixels(p, data->linear.x1, ox, sw);
		y1 = wxm_nsvg__convertToPixels(p, data->linear.y1, oy, sh);
		x2 = wxm_nsvg__convertToPixels(p, data->linear.x2, ox, sw);
		y2 = wxm_nsvg__convertToPixels(p, data->linear.y2, oy, sh);
		// Calculate transform aligned to the line
		dx = x2 - x1;
		dy = y2 - y1;
		grad->xform[0] = dy; grad->xform[1] = -dx;
		grad->xform[2] = dx; grad->xform[3] = dy;
		grad->xform[4] = x1; grad->xform[5] = y1;
	} else {
		float cx, cy, fx, fy, r;
		cx = wxm_nsvg__convertToPixels(p, data->radial.cx, ox, sw);
		cy = wxm_nsvg__convertToPixels(p, data->radial.cy, oy, sh);
		fx = wxm_nsvg__convertToPixels(p, data->radial.fx, ox, sw);
		fy = wxm_nsvg__convertToPixels(p, data->radial.fy, oy, sh);
		r = wxm_nsvg__convertToPixels(p, data->radial.r, 0, sl);
		// Calculate transform aligned to the circle
		grad->xform[0] = r; grad->xform[1] = 0;
		grad->xform[2] = 0; grad->xform[3] = r;
		grad->xform[4] = cx; grad->xform[5] = cy;
		grad->fx = fx / r;
		grad->fy = fy / r;
	}

	wxm_nsvg__xformMultiply(grad->xform, data->xform);
	wxm_nsvg__xformMultiply(grad->xform, xform);

	grad->spread = data->spread;
	memcpy(grad->stops, stops, nstops*sizeof(wxm_NSVGgradientStop));
	grad->nstops = nstops;

	*paintType = data->type;

	return grad;
}

static float wxm_nsvg__getAverageScale(float* t)
{
	float sx = sqrtf(t[0]*t[0] + t[2]*t[2]);
	float sy = sqrtf(t[1]*t[1] + t[3]*t[3]);
	return (sx + sy) * 0.5f;
}

static void wxm_nsvg__getLocalBounds(float* bounds, wxm_NSVGshape *shape, float* xform)
{
	wxm_NSVGpath* path;
	float curve[4*2], curveBounds[4];
	int i, first = 1;
	for (path = shape->paths; path != NULL; path = path->next) {
		wxm_nsvg__xformPoint(&curve[0], &curve[1], path->pts[0], path->pts[1], xform);
		for (i = 0; i < path->npts-1; i += 3) {
			wxm_nsvg__xformPoint(&curve[2], &curve[3], path->pts[(i+1)*2], path->pts[(i+1)*2+1], xform);
			wxm_nsvg__xformPoint(&curve[4], &curve[5], path->pts[(i+2)*2], path->pts[(i+2)*2+1], xform);
			wxm_nsvg__xformPoint(&curve[6], &curve[7], path->pts[(i+3)*2], path->pts[(i+3)*2+1], xform);
			wxm_nsvg__curveBounds(curveBounds, curve);
			if (first) {
				bounds[0] = curveBounds[0];
				bounds[1] = curveBounds[1];
				bounds[2] = curveBounds[2];
				bounds[3] = curveBounds[3];
				first = 0;
			} else {
				bounds[0] = wxm_nsvg__minf(bounds[0], curveBounds[0]);
				bounds[1] = wxm_nsvg__minf(bounds[1], curveBounds[1]);
				bounds[2] = wxm_nsvg__maxf(bounds[2], curveBounds[2]);
				bounds[3] = wxm_nsvg__maxf(bounds[3], curveBounds[3]);
			}
			curve[0] = curve[6];
			curve[1] = curve[7];
		}
	}
}

static void wxm_nsvg__addShape(wxm_NSVGparser* p)
{
	wxm_NSVGattrib* attr = wxm_nsvg__getAttr(p);
	float scale = 1.0f;
	wxm_NSVGshape* shape;
	wxm_NSVGpath* path;
	int i;

	if (p->plist == NULL)
		return;

	shape = (wxm_NSVGshape*)malloc(sizeof(wxm_NSVGshape));
	if (shape == NULL) goto error;
	memset(shape, 0, sizeof(wxm_NSVGshape));

	memcpy(shape->id, attr->id, sizeof shape->id);
	memcpy(shape->fillGradient, attr->fillGradient, sizeof shape->fillGradient);
	memcpy(shape->strokeGradient, attr->strokeGradient, sizeof shape->strokeGradient);
	memcpy(shape->xform, attr->xform, sizeof shape->xform);
	scale = wxm_nsvg__getAverageScale(attr->xform);
	shape->strokeWidth = attr->strokeWidth * scale;
	shape->strokeDashOffset = attr->strokeDashOffset * scale;
	shape->strokeDashCount = (char)attr->strokeDashCount;
	for (i = 0; i < attr->strokeDashCount; i++)
		shape->strokeDashArray[i] = attr->strokeDashArray[i] * scale;
	shape->strokeLineJoin = attr->strokeLineJoin;
	shape->strokeLineCap = attr->strokeLineCap;
	shape->miterLimit = attr->miterLimit;
	shape->fillRule = attr->fillRule;
	shape->opacity = attr->opacity;

	shape->paths = p->plist;
	p->plist = NULL;

	// Calculate shape bounds
	shape->bounds[0] = shape->paths->bounds[0];
	shape->bounds[1] = shape->paths->bounds[1];
	shape->bounds[2] = shape->paths->bounds[2];
	shape->bounds[3] = shape->paths->bounds[3];
	for (path = shape->paths->next; path != NULL; path = path->next) {
		shape->bounds[0] = wxm_nsvg__minf(shape->bounds[0], path->bounds[0]);
		shape->bounds[1] = wxm_nsvg__minf(shape->bounds[1], path->bounds[1]);
		shape->bounds[2] = wxm_nsvg__maxf(shape->bounds[2], path->bounds[2]);
		shape->bounds[3] = wxm_nsvg__maxf(shape->bounds[3], path->bounds[3]);
	}

	// Set fill
	if (attr->hasFill == 0) {
		shape->fill.type = wxm_NSVG_PAINT_NONE;
	} else if (attr->hasFill == 1) {
		shape->fill.type = wxm_NSVG_PAINT_COLOR;
		shape->fill.color = attr->fillColor;
		shape->fill.color |= (unsigned int)(attr->fillOpacity*255) << 24;
	} else if (attr->hasFill == 2) {
		shape->fill.type = wxm_NSVG_PAINT_UNDEF;
	}

	// Set stroke
	if (attr->hasStroke == 0) {
		shape->stroke.type = wxm_NSVG_PAINT_NONE;
	} else if (attr->hasStroke == 1) {
		shape->stroke.type = wxm_NSVG_PAINT_COLOR;
		shape->stroke.color = attr->strokeColor;
		shape->stroke.color |= (unsigned int)(attr->strokeOpacity*255) << 24;
	} else if (attr->hasStroke == 2) {
		shape->stroke.type = wxm_NSVG_PAINT_UNDEF;
	}

	// Set flags
	shape->flags = (attr->visible ? wxm_NSVG_FLAGS_VISIBLE : 0x00);

	// Add to tail
	if (p->image->shapes == NULL)
		p->image->shapes = shape;
	else
		p->shapesTail->next = shape;
	p->shapesTail = shape;

	return;

error:
	if (shape) free(shape);
}

static void wxm_nsvg__addPath(wxm_NSVGparser* p, char closed)
{
	wxm_NSVGattrib* attr = wxm_nsvg__getAttr(p);
	wxm_NSVGpath* path = NULL;
	float bounds[4];
	float* curve;
	int i;

	if (p->npts < 4)
		return;

	if (closed)
		wxm_nsvg__lineTo(p, p->pts[0], p->pts[1]);

	// Expect 1 + N*3 points (N = number of cubic bezier segments).
	if ((p->npts % 3) != 1)
		return;

	path = (wxm_NSVGpath*)malloc(sizeof(wxm_NSVGpath));
	if (path == NULL) goto error;
	memset(path, 0, sizeof(wxm_NSVGpath));

	path->pts = (float*)malloc(p->npts*2*sizeof(float));
	if (path->pts == NULL) goto error;
	path->closed = closed;
	path->npts = p->npts;

	// Transform path.
	for (i = 0; i < p->npts; ++i)
		wxm_nsvg__xformPoint(&path->pts[i*2], &path->pts[i*2+1], p->pts[i*2], p->pts[i*2+1], attr->xform);

	// Find bounds
	for (i = 0; i < path->npts-1; i += 3) {
		curve = &path->pts[i*2];
		wxm_nsvg__curveBounds(bounds, curve);
		if (i == 0) {
			path->bounds[0] = bounds[0];
			path->bounds[1] = bounds[1];
			path->bounds[2] = bounds[2];
			path->bounds[3] = bounds[3];
		} else {
			path->bounds[0] = wxm_nsvg__minf(path->bounds[0], bounds[0]);
			path->bounds[1] = wxm_nsvg__minf(path->bounds[1], bounds[1]);
			path->bounds[2] = wxm_nsvg__maxf(path->bounds[2], bounds[2]);
			path->bounds[3] = wxm_nsvg__maxf(path->bounds[3], bounds[3]);
		}
	}

	path->next = p->plist;
	p->plist = path;

	return;

error:
	if (path != NULL) {
		if (path->pts != NULL) free(path->pts);
		free(path);
	}
}

// We roll our own string to float because the std library one uses locale and messes things up.
static double wxm_nsvg__atof(const char* s)
{
	char* cur = (char*)s;
	char* end = NULL;
	double res = 0.0, sign = 1.0;
	long long intPart = 0, fracPart = 0;
	char hasIntPart = 0, hasFracPart = 0;

	// Parse optional sign
	if (*cur == '+') {
		cur++;
	} else if (*cur == '-') {
		sign = -1;
		cur++;
	}

	// Parse integer part
	if (wxm_nsvg__isdigit(*cur)) {
		// Parse digit sequence
		intPart = strtoll(cur, &end, 10);
		if (cur != end) {
			res = (double)intPart;
			hasIntPart = 1;
			cur = end;
		}
	}

	// Parse fractional part.
	if (*cur == '.') {
		cur++; // Skip '.'
		if (wxm_nsvg__isdigit(*cur)) {
			// Parse digit sequence
			fracPart = strtoll(cur, &end, 10);
			if (cur != end) {
				res += (double)fracPart / pow(10.0, (double)(end - cur));
				hasFracPart = 1;
				cur = end;
			}
		}
	}

	// A valid number should have integer or fractional part.
	if (!hasIntPart && !hasFracPart)
		return 0.0;

	// Parse optional exponent
	if (*cur == 'e' || *cur == 'E') {
		long expPart = 0;
		cur++; // skip 'E'
		expPart = strtol(cur, &end, 10); // Parse digit sequence with sign
		if (cur != end) {
			res *= pow(10.0, (double)expPart);
		}
	}

	return res * sign;
}


static const char* wxm_nsvg__parseNumber(const char* s, char* it, const int size)
{
	const int last = size-1;
	int i = 0;

	// sign
	if (*s == '-' || *s == '+') {
		if (i < last) it[i++] = *s;
		s++;
	}
	// integer part
	while (*s && wxm_nsvg__isdigit(*s)) {
		if (i < last) it[i++] = *s;
		s++;
	}
	if (*s == '.') {
		// decimal point
		if (i < last) it[i++] = *s;
		s++;
		// fraction part
		while (*s && wxm_nsvg__isdigit(*s)) {
			if (i < last) it[i++] = *s;
			s++;
		}
	}
	// exponent
	if ((*s == 'e' || *s == 'E') && (s[1] != 'm' && s[1] != 'x')) {
		if (i < last) it[i++] = *s;
		s++;
		if (*s == '-' || *s == '+') {
			if (i < last) it[i++] = *s;
			s++;
		}
		while (*s && wxm_nsvg__isdigit(*s)) {
			if (i < last) it[i++] = *s;
			s++;
		}
	}
	it[i] = '\0';

	return s;
}

static const char* wxm_nsvg__getNextPathItemWhenArcFlag(const char* s, char* it)
{
	it[0] = '\0';
	while (*s && (wxm_nsvg__isspace(*s) || *s == ',')) s++;
	if (!*s) return s;
	if (*s == '0' || *s == '1') {
		it[0] = *s++;
		it[1] = '\0';
		return s;
	}
	return s;
}

static const char* wxm_nsvg__getNextPathItem(const char* s, char* it)
{
	it[0] = '\0';
	// Skip white spaces and commas
	while (*s && (wxm_nsvg__isspace(*s) || *s == ',')) s++;
	if (!*s) return s;
	if (*s == '-' || *s == '+' || *s == '.' || wxm_nsvg__isdigit(*s)) {
		s = wxm_nsvg__parseNumber(s, it, 64);
	} else {
		// Parse command
		it[0] = *s++;
		it[1] = '\0';
		return s;
	}

	return s;
}

static unsigned int wxm_nsvg__parseColorHex(const char* str)
{
	unsigned int r=0, g=0, b=0;
	if (sscanf(str, "#%2x%2x%2x", &r, &g, &b) == 3 )		// 2 digit hex
		return wxm_NSVG_RGB(r, g, b);
	if (sscanf(str, "#%1x%1x%1x", &r, &g, &b) == 3 )		// 1 digit hex, e.g. #abc -> 0xccbbaa
		return wxm_NSVG_RGB(r*17, g*17, b*17);			// same effect as (r<<4|r), (g<<4|g), ..
	return wxm_NSVG_RGB(128, 128, 128);
}

// Parse rgb color. The pointer 'str' must point at "rgb(" (4+ characters).
// This function returns gray (rgb(128, 128, 128) == '#808080') on parse errors
// for backwards compatibility. Note: other image viewers return black instead.

static unsigned int wxm_nsvg__parseColorRGB(const char* str)
{
	int i;
	unsigned int rgbi[3];
	float rgbf[3];
	// try decimal integers first
	if (sscanf(str, "rgb(%u, %u, %u)", &rgbi[0], &rgbi[1], &rgbi[2]) != 3) {
		// integers failed, try percent values (float, locale independent)
		const char delimiter[3] = {',', ',', ')'};
		str += 4; // skip "rgb("
		for (i = 0; i < 3; i++) {
			while (*str && (wxm_nsvg__isspace(*str))) str++; 	// skip leading spaces
			if (*str == '+') str++;				// skip '+' (don't allow '-')
			if (!*str) break;
			rgbf[i] = wxm_nsvg__atof(str);

			// Note 1: it would be great if wxm_nsvg__atof() returned how many
			// bytes it consumed but it doesn't. We need to skip the number,
			// the '%' character, spaces, and the delimiter ',' or ')'.

			// Note 2: The following code does not allow values like "33.%",
			// i.e. a decimal point w/o fractional part, but this is consistent
			// with other image viewers, e.g. firefox, chrome, eog, gimp.

			while (*str && wxm_nsvg__isdigit(*str)) str++;		// skip integer part
			if (*str == '.') {
				str++;
				if (!wxm_nsvg__isdigit(*str)) break;		// error: no digit after '.'
				while (*str && wxm_nsvg__isdigit(*str)) str++;	// skip fractional part
			}
			if (*str == '%') str++; else break;
			while (wxm_nsvg__isspace(*str)) str++;
			if (*str == delimiter[i]) str++;
			else break;
		}
		if (i == 3) {
			rgbi[0] = roundf(rgbf[0] * 2.55f);
			rgbi[1] = roundf(rgbf[1] * 2.55f);
			rgbi[2] = roundf(rgbf[2] * 2.55f);
		} else {
			rgbi[0] = rgbi[1] = rgbi[2] = 128;
		}
	}
	// clip values as the CSS spec requires
	for (i = 0; i < 3; i++) {
		if (rgbi[i] > 255) rgbi[i] = 255;
	}
	return wxm_NSVG_RGB(rgbi[0], rgbi[1], rgbi[2]);
}

typedef struct wxm_NSVGNamedColor {
	const char* name;
	unsigned int color;
} wxm_NSVGNamedColor;

wxm_NSVGNamedColor wxm_nsvg__colors[] = {

	{ "red", wxm_NSVG_RGB(255, 0, 0) },
	{ "green", wxm_NSVG_RGB( 0, 128, 0) },
	{ "blue", wxm_NSVG_RGB( 0, 0, 255) },
	{ "yellow", wxm_NSVG_RGB(255, 255, 0) },
	{ "cyan", wxm_NSVG_RGB( 0, 255, 255) },
	{ "magenta", wxm_NSVG_RGB(255, 0, 255) },
	{ "black", wxm_NSVG_RGB( 0, 0, 0) },
	{ "grey", wxm_NSVG_RGB(128, 128, 128) },
	{ "gray", wxm_NSVG_RGB(128, 128, 128) },
	{ "white", wxm_NSVG_RGB(255, 255, 255) },

#ifdef NANOSVG_ALL_COLOR_KEYWORDS
	{ "aliceblue", wxm_NSVG_RGB(240, 248, 255) },
	{ "antiquewhite", wxm_NSVG_RGB(250, 235, 215) },
	{ "aqua", wxm_NSVG_RGB( 0, 255, 255) },
	{ "aquamarine", wxm_NSVG_RGB(127, 255, 212) },
	{ "azure", wxm_NSVG_RGB(240, 255, 255) },
	{ "beige", wxm_NSVG_RGB(245, 245, 220) },
	{ "bisque", wxm_NSVG_RGB(255, 228, 196) },
	{ "blanchedalmond", wxm_NSVG_RGB(255, 235, 205) },
	{ "blueviolet", wxm_NSVG_RGB(138, 43, 226) },
	{ "brown", wxm_NSVG_RGB(165, 42, 42) },
	{ "burlywood", wxm_NSVG_RGB(222, 184, 135) },
	{ "cadetblue", wxm_NSVG_RGB( 95, 158, 160) },
	{ "chartreuse", wxm_NSVG_RGB(127, 255, 0) },
	{ "chocolate", wxm_NSVG_RGB(210, 105, 30) },
	{ "coral", wxm_NSVG_RGB(255, 127, 80) },
	{ "cornflowerblue", wxm_NSVG_RGB(100, 149, 237) },
	{ "cornsilk", wxm_NSVG_RGB(255, 248, 220) },
	{ "crimson", wxm_NSVG_RGB(220, 20, 60) },
	{ "darkblue", wxm_NSVG_RGB( 0, 0, 139) },
	{ "darkcyan", wxm_NSVG_RGB( 0, 139, 139) },
	{ "darkgoldenrod", wxm_NSVG_RGB(184, 134, 11) },
	{ "darkgray", wxm_NSVG_RGB(169, 169, 169) },
	{ "darkgreen", wxm_NSVG_RGB( 0, 100, 0) },
	{ "darkgrey", wxm_NSVG_RGB(169, 169, 169) },
	{ "darkkhaki", wxm_NSVG_RGB(189, 183, 107) },
	{ "darkmagenta", wxm_NSVG_RGB(139, 0, 139) },
	{ "darkolivegreen", wxm_NSVG_RGB( 85, 107, 47) },
	{ "darkorange", wxm_NSVG_RGB(255, 140, 0) },
	{ "darkorchid", wxm_NSVG_RGB(153, 50, 204) },
	{ "darkred", wxm_NSVG_RGB(139, 0, 0) },
	{ "darksalmon", wxm_NSVG_RGB(233, 150, 122) },
	{ "darkseagreen", wxm_NSVG_RGB(143, 188, 143) },
	{ "darkslateblue", wxm_NSVG_RGB( 72, 61, 139) },
	{ "darkslategray", wxm_NSVG_RGB( 47, 79, 79) },
	{ "darkslategrey", wxm_NSVG_RGB( 47, 79, 79) },
	{ "darkturquoise", wxm_NSVG_RGB( 0, 206, 209) },
	{ "darkviolet", wxm_NSVG_RGB(148, 0, 211) },
	{ "deeppink", wxm_NSVG_RGB(255, 20, 147) },
	{ "deepskyblue", wxm_NSVG_RGB( 0, 191, 255) },
	{ "dimgray", wxm_NSVG_RGB(105, 105, 105) },
	{ "dimgrey", wxm_NSVG_RGB(105, 105, 105) },
	{ "dodgerblue", wxm_NSVG_RGB( 30, 144, 255) },
	{ "firebrick", wxm_NSVG_RGB(178, 34, 34) },
	{ "floralwhite", wxm_NSVG_RGB(255, 250, 240) },
	{ "forestgreen", wxm_NSVG_RGB( 34, 139, 34) },
	{ "fuchsia", wxm_NSVG_RGB(255, 0, 255) },
	{ "gainsboro", wxm_NSVG_RGB(220, 220, 220) },
	{ "ghostwhite", wxm_NSVG_RGB(248, 248, 255) },
	{ "gold", wxm_NSVG_RGB(255, 215, 0) },
	{ "goldenrod", wxm_NSVG_RGB(218, 165, 32) },
	{ "greenyellow", wxm_NSVG_RGB(173, 255, 47) },
	{ "honeydew", wxm_NSVG_RGB(240, 255, 240) },
	{ "hotpink", wxm_NSVG_RGB(255, 105, 180) },
	{ "indianred", wxm_NSVG_RGB(205, 92, 92) },
	{ "indigo", wxm_NSVG_RGB( 75, 0, 130) },
	{ "ivory", wxm_NSVG_RGB(255, 255, 240) },
	{ "khaki", wxm_NSVG_RGB(240, 230, 140) },
	{ "lavender", wxm_NSVG_RGB(230, 230, 250) },
	{ "lavenderblush", wxm_NSVG_RGB(255, 240, 245) },
	{ "lawngreen", wxm_NSVG_RGB(124, 252, 0) },
	{ "lemonchiffon", wxm_NSVG_RGB(255, 250, 205) },
	{ "lightblue", wxm_NSVG_RGB(173, 216, 230) },
	{ "lightcoral", wxm_NSVG_RGB(240, 128, 128) },
	{ "lightcyan", wxm_NSVG_RGB(224, 255, 255) },
	{ "lightgoldenrodyellow", wxm_NSVG_RGB(250, 250, 210) },
	{ "lightgray", wxm_NSVG_RGB(211, 211, 211) },
	{ "lightgreen", wxm_NSVG_RGB(144, 238, 144) },
	{ "lightgrey", wxm_NSVG_RGB(211, 211, 211) },
	{ "lightpink", wxm_NSVG_RGB(255, 182, 193) },
	{ "lightsalmon", wxm_NSVG_RGB(255, 160, 122) },
	{ "lightseagreen", wxm_NSVG_RGB( 32, 178, 170) },
	{ "lightskyblue", wxm_NSVG_RGB(135, 206, 250) },
	{ "lightslategray", wxm_NSVG_RGB(119, 136, 153) },
	{ "lightslategrey", wxm_NSVG_RGB(119, 136, 153) },
	{ "lightsteelblue", wxm_NSVG_RGB(176, 196, 222) },
	{ "lightyellow", wxm_NSVG_RGB(255, 255, 224) },
	{ "lime", wxm_NSVG_RGB( 0, 255, 0) },
	{ "limegreen", wxm_NSVG_RGB( 50, 205, 50) },
	{ "linen", wxm_NSVG_RGB(250, 240, 230) },
	{ "maroon", wxm_NSVG_RGB(128, 0, 0) },
	{ "mediumaquamarine", wxm_NSVG_RGB(102, 205, 170) },
	{ "mediumblue", wxm_NSVG_RGB( 0, 0, 205) },
	{ "mediumorchid", wxm_NSVG_RGB(186, 85, 211) },
	{ "mediumpurple", wxm_NSVG_RGB(147, 112, 219) },
	{ "mediumseagreen", wxm_NSVG_RGB( 60, 179, 113) },
	{ "mediumslateblue", wxm_NSVG_RGB(123, 104, 238) },
	{ "mediumspringgreen", wxm_NSVG_RGB( 0, 250, 154) },
	{ "mediumturquoise", wxm_NSVG_RGB( 72, 209, 204) },
	{ "mediumvioletred", wxm_NSVG_RGB(199, 21, 133) },
	{ "midnightblue", wxm_NSVG_RGB( 25, 25, 112) },
	{ "mintcream", wxm_NSVG_RGB(245, 255, 250) },
	{ "mistyrose", wxm_NSVG_RGB(255, 228, 225) },
	{ "moccasin", wxm_NSVG_RGB(255, 228, 181) },
	{ "navajowhite", wxm_NSVG_RGB(255, 222, 173) },
	{ "navy", wxm_NSVG_RGB( 0, 0, 128) },
	{ "oldlace", wxm_NSVG_RGB(253, 245, 230) },
	{ "olive", wxm_NSVG_RGB(128, 128, 0) },
	{ "olivedrab", wxm_NSVG_RGB(107, 142, 35) },
	{ "orange", wxm_NSVG_RGB(255, 165, 0) },
	{ "orangered", wxm_NSVG_RGB(255, 69, 0) },
	{ "orchid", wxm_NSVG_RGB(218, 112, 214) },
	{ "palegoldenrod", wxm_NSVG_RGB(238, 232, 170) },
	{ "palegreen", wxm_NSVG_RGB(152, 251, 152) },
	{ "paleturquoise", wxm_NSVG_RGB(175, 238, 238) },
	{ "palevioletred", wxm_NSVG_RGB(219, 112, 147) },
	{ "papayawhip", wxm_NSVG_RGB(255, 239, 213) },
	{ "peachpuff", wxm_NSVG_RGB(255, 218, 185) },
	{ "peru", wxm_NSVG_RGB(205, 133, 63) },
	{ "pink", wxm_NSVG_RGB(255, 192, 203) },
	{ "plum", wxm_NSVG_RGB(221, 160, 221) },
	{ "powderblue", wxm_NSVG_RGB(176, 224, 230) },
	{ "purple", wxm_NSVG_RGB(128, 0, 128) },
	{ "rosybrown", wxm_NSVG_RGB(188, 143, 143) },
	{ "royalblue", wxm_NSVG_RGB( 65, 105, 225) },
	{ "saddlebrown", wxm_NSVG_RGB(139, 69, 19) },
	{ "salmon", wxm_NSVG_RGB(250, 128, 114) },
	{ "sandybrown", wxm_NSVG_RGB(244, 164, 96) },
	{ "seagreen", wxm_NSVG_RGB( 46, 139, 87) },
	{ "seashell", wxm_NSVG_RGB(255, 245, 238) },
	{ "sienna", wxm_NSVG_RGB(160, 82, 45) },
	{ "silver", wxm_NSVG_RGB(192, 192, 192) },
	{ "skyblue", wxm_NSVG_RGB(135, 206, 235) },
	{ "slateblue", wxm_NSVG_RGB(106, 90, 205) },
	{ "slategray", wxm_NSVG_RGB(112, 128, 144) },
	{ "slategrey", wxm_NSVG_RGB(112, 128, 144) },
	{ "snow", wxm_NSVG_RGB(255, 250, 250) },
	{ "springgreen", wxm_NSVG_RGB( 0, 255, 127) },
	{ "steelblue", wxm_NSVG_RGB( 70, 130, 180) },
	{ "tan", wxm_NSVG_RGB(210, 180, 140) },
	{ "teal", wxm_NSVG_RGB( 0, 128, 128) },
	{ "thistle", wxm_NSVG_RGB(216, 191, 216) },
	{ "tomato", wxm_NSVG_RGB(255, 99, 71) },
	{ "turquoise", wxm_NSVG_RGB( 64, 224, 208) },
	{ "violet", wxm_NSVG_RGB(238, 130, 238) },
	{ "wheat", wxm_NSVG_RGB(245, 222, 179) },
	{ "whitesmoke", wxm_NSVG_RGB(245, 245, 245) },
	{ "yellowgreen", wxm_NSVG_RGB(154, 205, 50) },
#endif
};

static unsigned int wxm_nsvg__parseColorName(const char* str)
{
	int i, ncolors = sizeof(wxm_nsvg__colors) / sizeof(wxm_NSVGNamedColor);

	for (i = 0; i < ncolors; i++) {
		if (strcmp(wxm_nsvg__colors[i].name, str) == 0) {
			return wxm_nsvg__colors[i].color;
		}
	}

	return wxm_NSVG_RGB(128, 128, 128);
}

static unsigned int wxm_nsvg__parseColor(const char* str)
{
	size_t len = 0;
	while(*str == ' ') ++str;
	len = strlen(str);
	if (len >= 1 && *str == '#')
		return wxm_nsvg__parseColorHex(str);
	else if (len >= 4 && str[0] == 'r' && str[1] == 'g' && str[2] == 'b' && str[3] == '(')
		return wxm_nsvg__parseColorRGB(str);
	return wxm_nsvg__parseColorName(str);
}

static float wxm_nsvg__parseOpacity(const char* str)
{
	float val = wxm_nsvg__atof(str);
	if (val < 0.0f) val = 0.0f;
	if (val > 1.0f) val = 1.0f;
	return val;
}

static float wxm_nsvg__parseMiterLimit(const char* str)
{
	float val = wxm_nsvg__atof(str);
	if (val < 0.0f) val = 0.0f;
	return val;
}

static int wxm_nsvg__parseUnits(const char* units)
{
	if (units[0] == 'p' && units[1] == 'x')
		return wxm_NSVG_UNITS_PX;
	else if (units[0] == 'p' && units[1] == 't')
		return wxm_NSVG_UNITS_PT;
	else if (units[0] == 'p' && units[1] == 'c')
		return wxm_NSVG_UNITS_PC;
	else if (units[0] == 'm' && units[1] == 'm')
		return wxm_NSVG_UNITS_MM;
	else if (units[0] == 'c' && units[1] == 'm')
		return wxm_NSVG_UNITS_CM;
	else if (units[0] == 'i' && units[1] == 'n')
		return wxm_NSVG_UNITS_IN;
	else if (units[0] == '%')
		return wxm_NSVG_UNITS_PERCENT;
	else if (units[0] == 'e' && units[1] == 'm')
		return wxm_NSVG_UNITS_EM;
	else if (units[0] == 'e' && units[1] == 'x')
		return wxm_NSVG_UNITS_EX;
	return wxm_NSVG_UNITS_USER;
}

static int wxm_nsvg__isCoordinate(const char* s)
{
	// optional sign
	if (*s == '-' || *s == '+')
		s++;
	// must have at least one digit, or start by a dot
	return (wxm_nsvg__isdigit(*s) || *s == '.');
}

static wxm_NSVGcoordinate wxm_nsvg__parseCoordinateRaw(const char* str)
{
	wxm_NSVGcoordinate coord = {0, wxm_NSVG_UNITS_USER};
	char buf[64];
	coord.units = wxm_nsvg__parseUnits(wxm_nsvg__parseNumber(str, buf, 64));
	coord.value = wxm_nsvg__atof(buf);
	return coord;
}

static wxm_NSVGcoordinate wxm_nsvg__coord(float v, int units)
{
	wxm_NSVGcoordinate coord = {v, units};
	return coord;
}

static float wxm_nsvg__parseCoordinate(wxm_NSVGparser* p, const char* str, float orig, float length)
{
	wxm_NSVGcoordinate coord = wxm_nsvg__parseCoordinateRaw(str);
	return wxm_nsvg__convertToPixels(p, coord, orig, length);
}

static int wxm_nsvg__parseTransformArgs(const char* str, float* args, int maxNa, int* na)
{
	const char* end;
	const char* ptr;
	char it[64];

	*na = 0;
	ptr = str;
	while (*ptr && *ptr != '(') ++ptr;
	if (*ptr == 0)
		return 1;
	end = ptr;
	while (*end && *end != ')') ++end;
	if (*end == 0)
		return 1;

	while (ptr < end) {
		if (*ptr == '-' || *ptr == '+' || *ptr == '.' || wxm_nsvg__isdigit(*ptr)) {
			if (*na >= maxNa) return 0;
			ptr = wxm_nsvg__parseNumber(ptr, it, 64);
			args[(*na)++] = (float)wxm_nsvg__atof(it);
		} else {
			++ptr;
		}
	}
	return (int)(end - str);
}


static int wxm_nsvg__parseMatrix(float* xform, const char* str)
{
	float t[6];
	int na = 0;
	int len = wxm_nsvg__parseTransformArgs(str, t, 6, &na);
	if (na != 6) return len;
	memcpy(xform, t, sizeof(float)*6);
	return len;
}

static int wxm_nsvg__parseTranslate(float* xform, const char* str)
{
	float args[2];
	float t[6];
	int na = 0;
	int len = wxm_nsvg__parseTransformArgs(str, args, 2, &na);
	if (na == 1) args[1] = 0.0;

	wxm_nsvg__xformSetTranslation(t, args[0], args[1]);
	memcpy(xform, t, sizeof(float)*6);
	return len;
}

static int wxm_nsvg__parseScale(float* xform, const char* str)
{
	float args[2];
	int na = 0;
	float t[6];
	int len = wxm_nsvg__parseTransformArgs(str, args, 2, &na);
	if (na == 1) args[1] = args[0];
	wxm_nsvg__xformSetScale(t, args[0], args[1]);
	memcpy(xform, t, sizeof(float)*6);
	return len;
}

static int wxm_nsvg__parseSkewX(float* xform, const char* str)
{
	float args[1];
	int na = 0;
	float t[6];
	int len = wxm_nsvg__parseTransformArgs(str, args, 1, &na);
	wxm_nsvg__xformSetSkewX(t, args[0]/180.0f*wxm_NSVG_PI);
	memcpy(xform, t, sizeof(float)*6);
	return len;
}

static int wxm_nsvg__parseSkewY(float* xform, const char* str)
{
	float args[1];
	int na = 0;
	float t[6];
	int len = wxm_nsvg__parseTransformArgs(str, args, 1, &na);
	wxm_nsvg__xformSetSkewY(t, args[0]/180.0f*wxm_NSVG_PI);
	memcpy(xform, t, sizeof(float)*6);
	return len;
}

static int wxm_nsvg__parseRotate(float* xform, const char* str)
{
	float args[3];
	int na = 0;
	float m[6];
	float t[6];
	int len = wxm_nsvg__parseTransformArgs(str, args, 3, &na);
	if (na == 1)
		args[1] = args[2] = 0.0f;
	wxm_nsvg__xformIdentity(m);

	if (na > 1) {
		wxm_nsvg__xformSetTranslation(t, -args[1], -args[2]);
		wxm_nsvg__xformMultiply(m, t);
	}

	wxm_nsvg__xformSetRotation(t, args[0]/180.0f*wxm_NSVG_PI);
	wxm_nsvg__xformMultiply(m, t);

	if (na > 1) {
		wxm_nsvg__xformSetTranslation(t, args[1], args[2]);
		wxm_nsvg__xformMultiply(m, t);
	}

	memcpy(xform, m, sizeof(float)*6);

	return len;
}

static void wxm_nsvg__parseTransform(float* xform, const char* str)
{
	float t[6];
	int len;
	wxm_nsvg__xformIdentity(xform);
	while (*str)
	{
		if (strncmp(str, "matrix", 6) == 0)
			len = wxm_nsvg__parseMatrix(t, str);
		else if (strncmp(str, "translate", 9) == 0)
			len = wxm_nsvg__parseTranslate(t, str);
		else if (strncmp(str, "scale", 5) == 0)
			len = wxm_nsvg__parseScale(t, str);
		else if (strncmp(str, "rotate", 6) == 0)
			len = wxm_nsvg__parseRotate(t, str);
		else if (strncmp(str, "skewX", 5) == 0)
			len = wxm_nsvg__parseSkewX(t, str);
		else if (strncmp(str, "skewY", 5) == 0)
			len = wxm_nsvg__parseSkewY(t, str);
		else{
			++str;
			continue;
		}
		if (len != 0) {
			str += len;
		} else {
			++str;
			continue;
		}

		wxm_nsvg__xformPremultiply(xform, t);
	}
}

static void wxm_nsvg__parseUrl(char* id, const char* str)
{
	int i = 0;
	str += 4; // "url(";
	if (*str && *str == '#')
		str++;
	while (i < 63 && *str && *str != ')') {
		id[i] = *str++;
		i++;
	}
	id[i] = '\0';
}

static char wxm_nsvg__parseLineCap(const char* str)
{
	if (strcmp(str, "butt") == 0)
		return wxm_NSVG_CAP_BUTT;
	else if (strcmp(str, "round") == 0)
		return wxm_NSVG_CAP_ROUND;
	else if (strcmp(str, "square") == 0)
		return wxm_NSVG_CAP_SQUARE;
	// TODO: handle inherit.
	return wxm_NSVG_CAP_BUTT;
}

static char wxm_nsvg__parseLineJoin(const char* str)
{
	if (strcmp(str, "miter") == 0)
		return wxm_NSVG_JOIN_MITER;
	else if (strcmp(str, "round") == 0)
		return wxm_NSVG_JOIN_ROUND;
	else if (strcmp(str, "bevel") == 0)
		return wxm_NSVG_JOIN_BEVEL;
	// TODO: handle inherit.
	return wxm_NSVG_JOIN_MITER;
}

static char wxm_nsvg__parseFillRule(const char* str)
{
	if (strcmp(str, "nonzero") == 0)
		return wxm_NSVG_FILLRULE_NONZERO;
	else if (strcmp(str, "evenodd") == 0)
		return wxm_NSVG_FILLRULE_EVENODD;
	// TODO: handle inherit.
	return wxm_NSVG_FILLRULE_NONZERO;
}

static const char* wxm_nsvg__getNextDashItem(const char* s, char* it)
{
	int n = 0;
	it[0] = '\0';
	// Skip white spaces and commas
	while (*s && (wxm_nsvg__isspace(*s) || *s == ',')) s++;
	// Advance until whitespace, comma or end.
	while (*s && (!wxm_nsvg__isspace(*s) && *s != ',')) {
		if (n < 63)
			it[n++] = *s;
		s++;
	}
	it[n++] = '\0';
	return s;
}

static int wxm_nsvg__parseStrokeDashArray(wxm_NSVGparser* p, const char* str, float* strokeDashArray)
{
	char item[64];
	int count = 0, i;
	float sum = 0.0f;

	// Handle "none"
	if (str[0] == 'n')
		return 0;

	// Parse dashes
	while (*str) {
		str = wxm_nsvg__getNextDashItem(str, item);
		if (!*item) break;
		if (count < wxm_NSVG_MAX_DASHES)
			strokeDashArray[count++] = fabsf(wxm_nsvg__parseCoordinate(p, item, 0.0f, wxm_nsvg__actualLength(p)));
	}

	for (i = 0; i < count; i++)
		sum += strokeDashArray[i];
	if (sum <= 1e-6f)
		count = 0;

	return count;
}

static void wxm_nsvg__parseStyle(wxm_NSVGparser* p, const char* str);

static int wxm_nsvg__parseAttr(wxm_NSVGparser* p, const char* name, const char* value)
{
	float xform[6];
	wxm_NSVGattrib* attr = wxm_nsvg__getAttr(p);
	if (!attr) return 0;

	if (strcmp(name, "style") == 0) {
		wxm_nsvg__parseStyle(p, value);
	} else if (strcmp(name, "display") == 0) {
		if (strcmp(value, "none") == 0)
			attr->visible = 0;
		// Don't reset ->visible on display:inline, one display:none hides the whole subtree

	} else if (strcmp(name, "fill") == 0) {
		if (strcmp(value, "none") == 0) {
			attr->hasFill = 0;
		} else if (strncmp(value, "url(", 4) == 0) {
			attr->hasFill = 2;
			wxm_nsvg__parseUrl(attr->fillGradient, value);
		} else {
			attr->hasFill = 1;
			attr->fillColor = wxm_nsvg__parseColor(value);
		}
	} else if (strcmp(name, "opacity") == 0) {
		attr->opacity = wxm_nsvg__parseOpacity(value);
	} else if (strcmp(name, "fill-opacity") == 0) {
		attr->fillOpacity = wxm_nsvg__parseOpacity(value);
	} else if (strcmp(name, "stroke") == 0) {
		if (strcmp(value, "none") == 0) {
			attr->hasStroke = 0;
		} else if (strncmp(value, "url(", 4) == 0) {
			attr->hasStroke = 2;
			wxm_nsvg__parseUrl(attr->strokeGradient, value);
		} else {
			attr->hasStroke = 1;
			attr->strokeColor = wxm_nsvg__parseColor(value);
		}
	} else if (strcmp(name, "stroke-width") == 0) {
		attr->strokeWidth = wxm_nsvg__parseCoordinate(p, value, 0.0f, wxm_nsvg__actualLength(p));
	} else if (strcmp(name, "stroke-dasharray") == 0) {
		attr->strokeDashCount = wxm_nsvg__parseStrokeDashArray(p, value, attr->strokeDashArray);
	} else if (strcmp(name, "stroke-dashoffset") == 0) {
		attr->strokeDashOffset = wxm_nsvg__parseCoordinate(p, value, 0.0f, wxm_nsvg__actualLength(p));
	} else if (strcmp(name, "stroke-opacity") == 0) {
		attr->strokeOpacity = wxm_nsvg__parseOpacity(value);
	} else if (strcmp(name, "stroke-linecap") == 0) {
		attr->strokeLineCap = wxm_nsvg__parseLineCap(value);
	} else if (strcmp(name, "stroke-linejoin") == 0) {
		attr->strokeLineJoin = wxm_nsvg__parseLineJoin(value);
	} else if (strcmp(name, "stroke-miterlimit") == 0) {
		attr->miterLimit = wxm_nsvg__parseMiterLimit(value);
	} else if (strcmp(name, "fill-rule") == 0) {
		attr->fillRule = wxm_nsvg__parseFillRule(value);
	} else if (strcmp(name, "font-size") == 0) {
		attr->fontSize = wxm_nsvg__parseCoordinate(p, value, 0.0f, wxm_nsvg__actualLength(p));
	} else if (strcmp(name, "transform") == 0) {
		wxm_nsvg__parseTransform(xform, value);
		wxm_nsvg__xformPremultiply(attr->xform, xform);
	} else if (strcmp(name, "stop-color") == 0) {
		attr->stopColor = wxm_nsvg__parseColor(value);
	} else if (strcmp(name, "stop-opacity") == 0) {
		attr->stopOpacity = wxm_nsvg__parseOpacity(value);
	} else if (strcmp(name, "offset") == 0) {
		attr->stopOffset = wxm_nsvg__parseCoordinate(p, value, 0.0f, 1.0f);
	} else if (strcmp(name, "id") == 0) {
		strncpy(attr->id, value, 63);
		attr->id[63] = '\0';
	} else {
		return 0;
	}
	return 1;
}

static int wxm_nsvg__parseNameValue(wxm_NSVGparser* p, const char* start, const char* end)
{
	const char* str;
	const char* val;
	char name[512];
	char value[512];
	int n;

	str = start;
	while (str < end && *str != ':') ++str;

	val = str;

	// Right Trim
	while (str > start &&  (*str == ':' || wxm_nsvg__isspace(*str))) --str;
	++str;

	n = (int)(str - start);
	if (n > 511) n = 511;
	if (n) memcpy(name, start, n);
	name[n] = 0;

	while (val < end && (*val == ':' || wxm_nsvg__isspace(*val))) ++val;

	n = (int)(end - val);
	if (n > 511) n = 511;
	if (n) memcpy(value, val, n);
	value[n] = 0;

	return wxm_nsvg__parseAttr(p, name, value);
}

static void wxm_nsvg__parseStyle(wxm_NSVGparser* p, const char* str)
{
	const char* start;
	const char* end;

	while (*str) {
		// Left Trim
		while(*str && wxm_nsvg__isspace(*str)) ++str;
		start = str;
		while(*str && *str != ';') ++str;
		end = str;

		// Right Trim
		while (end > start &&  (*end == ';' || wxm_nsvg__isspace(*end))) --end;
		++end;

		wxm_nsvg__parseNameValue(p, start, end);
		if (*str) ++str;
	}
}

static void wxm_nsvg__parseAttribs(wxm_NSVGparser* p, const char** attr)
{
	int i;
	for (i = 0; attr[i]; i += 2)
	{
		if (strcmp(attr[i], "style") == 0)
			wxm_nsvg__parseStyle(p, attr[i + 1]);
		else
			wxm_nsvg__parseAttr(p, attr[i], attr[i + 1]);
	}
}

static int wxm_nsvg__getArgsPerElement(char cmd)
{
	switch (cmd) {
		case 'v':
		case 'V':
		case 'h':
		case 'H':
			return 1;
		case 'm':
		case 'M':
		case 'l':
		case 'L':
		case 't':
		case 'T':
			return 2;
		case 'q':
		case 'Q':
		case 's':
		case 'S':
			return 4;
		case 'c':
		case 'C':
			return 6;
		case 'a':
		case 'A':
			return 7;
		case 'z':
		case 'Z':
			return 0;
	}
	return -1;
}

static void wxm_nsvg__pathMoveTo(wxm_NSVGparser* p, float* cpx, float* cpy, float* args, int rel)
{
	if (rel) {
		*cpx += args[0];
		*cpy += args[1];
	} else {
		*cpx = args[0];
		*cpy = args[1];
	}
	wxm_nsvg__moveTo(p, *cpx, *cpy);
}

static void wxm_nsvg__pathLineTo(wxm_NSVGparser* p, float* cpx, float* cpy, float* args, int rel)
{
	if (rel) {
		*cpx += args[0];
		*cpy += args[1];
	} else {
		*cpx = args[0];
		*cpy = args[1];
	}
	wxm_nsvg__lineTo(p, *cpx, *cpy);
}

static void wxm_nsvg__pathHLineTo(wxm_NSVGparser* p, float* cpx, float* cpy, float* args, int rel)
{
	if (rel)
		*cpx += args[0];
	else
		*cpx = args[0];
	wxm_nsvg__lineTo(p, *cpx, *cpy);
}

static void wxm_nsvg__pathVLineTo(wxm_NSVGparser* p, float* cpx, float* cpy, float* args, int rel)
{
	if (rel)
		*cpy += args[0];
	else
		*cpy = args[0];
	wxm_nsvg__lineTo(p, *cpx, *cpy);
}

static void wxm_nsvg__pathCubicBezTo(wxm_NSVGparser* p, float* cpx, float* cpy,
								 float* cpx2, float* cpy2, float* args, int rel)
{
	float x2, y2, cx1, cy1, cx2, cy2;

	if (rel) {
		cx1 = *cpx + args[0];
		cy1 = *cpy + args[1];
		cx2 = *cpx + args[2];
		cy2 = *cpy + args[3];
		x2 = *cpx + args[4];
		y2 = *cpy + args[5];
	} else {
		cx1 = args[0];
		cy1 = args[1];
		cx2 = args[2];
		cy2 = args[3];
		x2 = args[4];
		y2 = args[5];
	}

	wxm_nsvg__cubicBezTo(p, cx1,cy1, cx2,cy2, x2,y2);

	*cpx2 = cx2;
	*cpy2 = cy2;
	*cpx = x2;
	*cpy = y2;
}

static void wxm_nsvg__pathCubicBezShortTo(wxm_NSVGparser* p, float* cpx, float* cpy,
									  float* cpx2, float* cpy2, float* args, int rel)
{
	float x1, y1, x2, y2, cx1, cy1, cx2, cy2;

	x1 = *cpx;
	y1 = *cpy;
	if (rel) {
		cx2 = *cpx + args[0];
		cy2 = *cpy + args[1];
		x2 = *cpx + args[2];
		y2 = *cpy + args[3];
	} else {
		cx2 = args[0];
		cy2 = args[1];
		x2 = args[2];
		y2 = args[3];
	}

	cx1 = 2*x1 - *cpx2;
	cy1 = 2*y1 - *cpy2;

	wxm_nsvg__cubicBezTo(p, cx1,cy1, cx2,cy2, x2,y2);

	*cpx2 = cx2;
	*cpy2 = cy2;
	*cpx = x2;
	*cpy = y2;
}

static void wxm_nsvg__pathQuadBezTo(wxm_NSVGparser* p, float* cpx, float* cpy,
								float* cpx2, float* cpy2, float* args, int rel)
{
	float x1, y1, x2, y2, cx, cy;
	float cx1, cy1, cx2, cy2;

	x1 = *cpx;
	y1 = *cpy;
	if (rel) {
		cx = *cpx + args[0];
		cy = *cpy + args[1];
		x2 = *cpx + args[2];
		y2 = *cpy + args[3];
	} else {
		cx = args[0];
		cy = args[1];
		x2 = args[2];
		y2 = args[3];
	}

	// Convert to cubic bezier
	cx1 = x1 + 2.0f/3.0f*(cx - x1);
	cy1 = y1 + 2.0f/3.0f*(cy - y1);
	cx2 = x2 + 2.0f/3.0f*(cx - x2);
	cy2 = y2 + 2.0f/3.0f*(cy - y2);

	wxm_nsvg__cubicBezTo(p, cx1,cy1, cx2,cy2, x2,y2);

	*cpx2 = cx;
	*cpy2 = cy;
	*cpx = x2;
	*cpy = y2;
}

static void wxm_nsvg__pathQuadBezShortTo(wxm_NSVGparser* p, float* cpx, float* cpy,
									 float* cpx2, float* cpy2, float* args, int rel)
{
	float x1, y1, x2, y2, cx, cy;
	float cx1, cy1, cx2, cy2;

	x1 = *cpx;
	y1 = *cpy;
	if (rel) {
		x2 = *cpx + args[0];
		y2 = *cpy + args[1];
	} else {
		x2 = args[0];
		y2 = args[1];
	}

	cx = 2*x1 - *cpx2;
	cy = 2*y1 - *cpy2;

	// Convert to cubix bezier
	cx1 = x1 + 2.0f/3.0f*(cx - x1);
	cy1 = y1 + 2.0f/3.0f*(cy - y1);
	cx2 = x2 + 2.0f/3.0f*(cx - x2);
	cy2 = y2 + 2.0f/3.0f*(cy - y2);

	wxm_nsvg__cubicBezTo(p, cx1,cy1, cx2,cy2, x2,y2);

	*cpx2 = cx;
	*cpy2 = cy;
	*cpx = x2;
	*cpy = y2;
}

static float wxm_nsvg__sqr(float x) { return x*x; }
static float wxm_nsvg__vmag(float x, float y) { return sqrtf(x*x + y*y); }

static float wxm_nsvg__vecrat(float ux, float uy, float vx, float vy)
{
	return (ux*vx + uy*vy) / (wxm_nsvg__vmag(ux,uy) * wxm_nsvg__vmag(vx,vy));
}

static float wxm_nsvg__vecang(float ux, float uy, float vx, float vy)
{
	float r = wxm_nsvg__vecrat(ux,uy, vx,vy);
	if (r < -1.0f) r = -1.0f;
	if (r > 1.0f) r = 1.0f;
	return ((ux*vy < uy*vx) ? -1.0f : 1.0f) * acosf(r);
}

static void wxm_nsvg__pathArcTo(wxm_NSVGparser* p, float* cpx, float* cpy, float* args, int rel)
{
	// Ported from canvg (https://code.google.com/p/canvg/)
	float rx, ry, rotx;
	float x1, y1, x2, y2, cx, cy, dx, dy, d;
	float x1p, y1p, cxp, cyp, s, sa, sb;
	float ux, uy, vx, vy, a1, da;
	float x, y, tanx, tany, a, px = 0, py = 0, ptanx = 0, ptany = 0, t[6];
	float sinrx, cosrx;
	int fa, fs;
	int i, ndivs;
	float hda, kappa;

	rx = fabsf(args[0]);				// y radius
	ry = fabsf(args[1]);				// x radius
	rotx = args[2] / 180.0f * wxm_NSVG_PI;		// x rotation angle
	fa = fabsf(args[3]) > 1e-6 ? 1 : 0;	// Large arc
	fs = fabsf(args[4]) > 1e-6 ? 1 : 0;	// Sweep direction
	x1 = *cpx;							// start point
	y1 = *cpy;
	if (rel) {							// end point
		x2 = *cpx + args[5];
		y2 = *cpy + args[6];
	} else {
		x2 = args[5];
		y2 = args[6];
	}

	dx = x1 - x2;
	dy = y1 - y2;
	d = sqrtf(dx*dx + dy*dy);
	if (d < 1e-6f || rx < 1e-6f || ry < 1e-6f) {
		// The arc degenerates to a line
		wxm_nsvg__lineTo(p, x2, y2);
		*cpx = x2;
		*cpy = y2;
		return;
	}

	sinrx = sinf(rotx);
	cosrx = cosf(rotx);

	// Convert to center point parameterization.
	// http://www.w3.org/TR/SVG11/implnote.html#ArcImplementationNotes
	// 1) Compute x1', y1'
	x1p = cosrx * dx / 2.0f + sinrx * dy / 2.0f;
	y1p = -sinrx * dx / 2.0f + cosrx * dy / 2.0f;
	d = wxm_nsvg__sqr(x1p)/wxm_nsvg__sqr(rx) + wxm_nsvg__sqr(y1p)/wxm_nsvg__sqr(ry);
	if (d > 1) {
		d = sqrtf(d);
		rx *= d;
		ry *= d;
	}
	// 2) Compute cx', cy'
	s = 0.0f;
	sa = wxm_nsvg__sqr(rx)*wxm_nsvg__sqr(ry) - wxm_nsvg__sqr(rx)*wxm_nsvg__sqr(y1p) - wxm_nsvg__sqr(ry)*wxm_nsvg__sqr(x1p);
	sb = wxm_nsvg__sqr(rx)*wxm_nsvg__sqr(y1p) + wxm_nsvg__sqr(ry)*wxm_nsvg__sqr(x1p);
	if (sa < 0.0f) sa = 0.0f;
	if (sb > 0.0f)
		s = sqrtf(sa / sb);
	if (fa == fs)
		s = -s;
	cxp = s * rx * y1p / ry;
	cyp = s * -ry * x1p / rx;

	// 3) Compute cx,cy from cx',cy'
	cx = (x1 + x2)/2.0f + cosrx*cxp - sinrx*cyp;
	cy = (y1 + y2)/2.0f + sinrx*cxp + cosrx*cyp;

	// 4) Calculate theta1, and delta theta.
	ux = (x1p - cxp) / rx;
	uy = (y1p - cyp) / ry;
	vx = (-x1p - cxp) / rx;
	vy = (-y1p - cyp) / ry;
	a1 = wxm_nsvg__vecang(1.0f,0.0f, ux,uy);	// Initial angle
	da = wxm_nsvg__vecang(ux,uy, vx,vy);		// Delta angle

//	if (vecrat(ux,uy,vx,vy) <= -1.0f) da = wxm_NSVG_PI;
//	if (vecrat(ux,uy,vx,vy) >= 1.0f) da = 0;

	if (fs == 0 && da > 0)
		da -= 2 * wxm_NSVG_PI;
	else if (fs == 1 && da < 0)
		da += 2 * wxm_NSVG_PI;

	// Approximate the arc using cubic spline segments.
	t[0] = cosrx; t[1] = sinrx;
	t[2] = -sinrx; t[3] = cosrx;
	t[4] = cx; t[5] = cy;

	// Split arc into max 90 degree segments.
	// The loop assumes an iteration per end point (including start and end), this +1.
	ndivs = (int)(fabsf(da) / (wxm_NSVG_PI*0.5f) + 1.0f);
	hda = (da / (float)ndivs) / 2.0f;
	// Fix for ticket #179: division by 0: avoid cotangens around 0 (infinite)
	if ((hda < 1e-3f) && (hda > -1e-3f))
		hda *= 0.5f;
	else
		hda = (1.0f - cosf(hda)) / sinf(hda);
	kappa = fabsf(4.0f / 3.0f * hda);
	if (da < 0.0f)
		kappa = -kappa;

	for (i = 0; i <= ndivs; i++) {
		a = a1 + da * ((float)i/(float)ndivs);
		dx = cosf(a);
		dy = sinf(a);
		wxm_nsvg__xformPoint(&x, &y, dx*rx, dy*ry, t); // position
		wxm_nsvg__xformVec(&tanx, &tany, -dy*rx * kappa, dx*ry * kappa, t); // tangent
		if (i > 0)
			wxm_nsvg__cubicBezTo(p, px+ptanx,py+ptany, x-tanx, y-tany, x, y);
		px = x;
		py = y;
		ptanx = tanx;
		ptany = tany;
	}

	*cpx = x2;
	*cpy = y2;
}

static void wxm_nsvg__parsePath(wxm_NSVGparser* p, const char** attr)
{
	const char* s = NULL;
	char cmd = '\0';
	float args[10];
	int nargs;
	int rargs = 0;
	char initPoint;
	float cpx, cpy, cpx2, cpy2;
	const char* tmp[4];
	char closedFlag;
	int i;
	char item[64];

	for (i = 0; attr[i]; i += 2) {
		if (strcmp(attr[i], "d") == 0) {
			s = attr[i + 1];
		} else {
			tmp[0] = attr[i];
			tmp[1] = attr[i + 1];
			tmp[2] = 0;
			tmp[3] = 0;
			wxm_nsvg__parseAttribs(p, tmp);
		}
	}

	if (s) {
		wxm_nsvg__resetPath(p);
		cpx = 0; cpy = 0;
		cpx2 = 0; cpy2 = 0;
		initPoint = 0;
		closedFlag = 0;
		nargs = 0;

		while (*s) {
			item[0] = '\0';
			if ((cmd == 'A' || cmd == 'a') && (nargs == 3 || nargs == 4))
				s = wxm_nsvg__getNextPathItemWhenArcFlag(s, item);
			if (!*item)
				s = wxm_nsvg__getNextPathItem(s, item);
			if (!*item) break;
			if (cmd != '\0' && wxm_nsvg__isCoordinate(item)) {
				if (nargs < 10)
					args[nargs++] = (float)wxm_nsvg__atof(item);
				if (nargs >= rargs) {
					switch (cmd) {
						case 'm':
						case 'M':
							wxm_nsvg__pathMoveTo(p, &cpx, &cpy, args, cmd == 'm' ? 1 : 0);
							// Moveto can be followed by multiple coordinate pairs,
							// which should be treated as linetos.
							cmd = (cmd == 'm') ? 'l' : 'L';
							rargs = wxm_nsvg__getArgsPerElement(cmd);
							cpx2 = cpx; cpy2 = cpy;
							initPoint = 1;
							break;
						case 'l':
						case 'L':
							wxm_nsvg__pathLineTo(p, &cpx, &cpy, args, cmd == 'l' ? 1 : 0);
							cpx2 = cpx; cpy2 = cpy;
							break;
						case 'H':
						case 'h':
							wxm_nsvg__pathHLineTo(p, &cpx, &cpy, args, cmd == 'h' ? 1 : 0);
							cpx2 = cpx; cpy2 = cpy;
							break;
						case 'V':
						case 'v':
							wxm_nsvg__pathVLineTo(p, &cpx, &cpy, args, cmd == 'v' ? 1 : 0);
							cpx2 = cpx; cpy2 = cpy;
							break;
						case 'C':
						case 'c':
							wxm_nsvg__pathCubicBezTo(p, &cpx, &cpy, &cpx2, &cpy2, args, cmd == 'c' ? 1 : 0);
							break;
						case 'S':
						case 's':
							wxm_nsvg__pathCubicBezShortTo(p, &cpx, &cpy, &cpx2, &cpy2, args, cmd == 's' ? 1 : 0);
							break;
						case 'Q':
						case 'q':
							wxm_nsvg__pathQuadBezTo(p, &cpx, &cpy, &cpx2, &cpy2, args, cmd == 'q' ? 1 : 0);
							break;
						case 'T':
						case 't':
							wxm_nsvg__pathQuadBezShortTo(p, &cpx, &cpy, &cpx2, &cpy2, args, cmd == 't' ? 1 : 0);
							break;
						case 'A':
						case 'a':
							wxm_nsvg__pathArcTo(p, &cpx, &cpy, args, cmd == 'a' ? 1 : 0);
							cpx2 = cpx; cpy2 = cpy;
							break;
						default:
							if (nargs >= 2) {
								cpx = args[nargs-2];
								cpy = args[nargs-1];
								cpx2 = cpx; cpy2 = cpy;
							}
							break;
					}
					nargs = 0;
				}
			} else {
				cmd = item[0];
				if (cmd == 'M' || cmd == 'm') {
					// Commit path.
					if (p->npts > 0)
						wxm_nsvg__addPath(p, closedFlag);
					// Start new subpath.
					wxm_nsvg__resetPath(p);
					closedFlag = 0;
					nargs = 0;
				} else if (initPoint == 0) {
					// Do not allow other commands until initial point has been set (moveTo called once).
					cmd = '\0';
				}
				if (cmd == 'Z' || cmd == 'z') {
					closedFlag = 1;
					// Commit path.
					if (p->npts > 0) {
						// Move current point to first point
						cpx = p->pts[0];
						cpy = p->pts[1];
						cpx2 = cpx; cpy2 = cpy;
						wxm_nsvg__addPath(p, closedFlag);
					}
					// Start new subpath.
					wxm_nsvg__resetPath(p);
					wxm_nsvg__moveTo(p, cpx, cpy);
					closedFlag = 0;
					nargs = 0;
				}
				rargs = wxm_nsvg__getArgsPerElement(cmd);
				if (rargs == -1) {
					// Command not recognized
					cmd = '\0';
					rargs = 0;
				}
			}
		}
		// Commit path.
		if (p->npts)
			wxm_nsvg__addPath(p, closedFlag);
	}

	wxm_nsvg__addShape(p);
}

static void wxm_nsvg__parseRect(wxm_NSVGparser* p, const char** attr)
{
	float x = 0.0f;
	float y = 0.0f;
	float w = 0.0f;
	float h = 0.0f;
	float rx = -1.0f; // marks not set
	float ry = -1.0f;
	int i;

	for (i = 0; attr[i]; i += 2) {
		if (!wxm_nsvg__parseAttr(p, attr[i], attr[i + 1])) {
			if (strcmp(attr[i], "x") == 0) x = wxm_nsvg__parseCoordinate(p, attr[i+1], wxm_nsvg__actualOrigX(p), wxm_nsvg__actualWidth(p));
			if (strcmp(attr[i], "y") == 0) y = wxm_nsvg__parseCoordinate(p, attr[i+1], wxm_nsvg__actualOrigY(p), wxm_nsvg__actualHeight(p));
			if (strcmp(attr[i], "width") == 0) w = wxm_nsvg__parseCoordinate(p, attr[i+1], 0.0f, wxm_nsvg__actualWidth(p));
			if (strcmp(attr[i], "height") == 0) h = wxm_nsvg__parseCoordinate(p, attr[i+1], 0.0f, wxm_nsvg__actualHeight(p));
			if (strcmp(attr[i], "rx") == 0) rx = fabsf(wxm_nsvg__parseCoordinate(p, attr[i+1], 0.0f, wxm_nsvg__actualWidth(p)));
			if (strcmp(attr[i], "ry") == 0) ry = fabsf(wxm_nsvg__parseCoordinate(p, attr[i+1], 0.0f, wxm_nsvg__actualHeight(p)));
		}
	}

	if (rx < 0.0f && ry > 0.0f) rx = ry;
	if (ry < 0.0f && rx > 0.0f) ry = rx;
	if (rx < 0.0f) rx = 0.0f;
	if (ry < 0.0f) ry = 0.0f;
	if (rx > w/2.0f) rx = w/2.0f;
	if (ry > h/2.0f) ry = h/2.0f;

	if (w != 0.0f && h != 0.0f) {
		wxm_nsvg__resetPath(p);

		if (rx < 0.00001f || ry < 0.0001f) {
			wxm_nsvg__moveTo(p, x, y);
			wxm_nsvg__lineTo(p, x+w, y);
			wxm_nsvg__lineTo(p, x+w, y+h);
			wxm_nsvg__lineTo(p, x, y+h);
		} else {
			// Rounded rectangle
			wxm_nsvg__moveTo(p, x+rx, y);
			wxm_nsvg__lineTo(p, x+w-rx, y);
			wxm_nsvg__cubicBezTo(p, x+w-rx*(1-wxm_NSVG_KAPPA90), y, x+w, y+ry*(1-wxm_NSVG_KAPPA90), x+w, y+ry);
			wxm_nsvg__lineTo(p, x+w, y+h-ry);
			wxm_nsvg__cubicBezTo(p, x+w, y+h-ry*(1-wxm_NSVG_KAPPA90), x+w-rx*(1-wxm_NSVG_KAPPA90), y+h, x+w-rx, y+h);
			wxm_nsvg__lineTo(p, x+rx, y+h);
			wxm_nsvg__cubicBezTo(p, x+rx*(1-wxm_NSVG_KAPPA90), y+h, x, y+h-ry*(1-wxm_NSVG_KAPPA90), x, y+h-ry);
			wxm_nsvg__lineTo(p, x, y+ry);
			wxm_nsvg__cubicBezTo(p, x, y+ry*(1-wxm_NSVG_KAPPA90), x+rx*(1-wxm_NSVG_KAPPA90), y, x+rx, y);
		}

		wxm_nsvg__addPath(p, 1);

		wxm_nsvg__addShape(p);
	}
}

static void wxm_nsvg__parseCircle(wxm_NSVGparser* p, const char** attr)
{
	float cx = 0.0f;
	float cy = 0.0f;
	float r = 0.0f;
	int i;

	for (i = 0; attr[i]; i += 2) {
		if (!wxm_nsvg__parseAttr(p, attr[i], attr[i + 1])) {
			if (strcmp(attr[i], "cx") == 0) cx = wxm_nsvg__parseCoordinate(p, attr[i+1], wxm_nsvg__actualOrigX(p), wxm_nsvg__actualWidth(p));
			if (strcmp(attr[i], "cy") == 0) cy = wxm_nsvg__parseCoordinate(p, attr[i+1], wxm_nsvg__actualOrigY(p), wxm_nsvg__actualHeight(p));
			if (strcmp(attr[i], "r") == 0) r = fabsf(wxm_nsvg__parseCoordinate(p, attr[i+1], 0.0f, wxm_nsvg__actualLength(p)));
		}
	}

	if (r > 0.0f) {
		wxm_nsvg__resetPath(p);

		wxm_nsvg__moveTo(p, cx+r, cy);
		wxm_nsvg__cubicBezTo(p, cx+r, cy+r*wxm_NSVG_KAPPA90, cx+r*wxm_NSVG_KAPPA90, cy+r, cx, cy+r);
		wxm_nsvg__cubicBezTo(p, cx-r*wxm_NSVG_KAPPA90, cy+r, cx-r, cy+r*wxm_NSVG_KAPPA90, cx-r, cy);
		wxm_nsvg__cubicBezTo(p, cx-r, cy-r*wxm_NSVG_KAPPA90, cx-r*wxm_NSVG_KAPPA90, cy-r, cx, cy-r);
		wxm_nsvg__cubicBezTo(p, cx+r*wxm_NSVG_KAPPA90, cy-r, cx+r, cy-r*wxm_NSVG_KAPPA90, cx+r, cy);

		wxm_nsvg__addPath(p, 1);

		wxm_nsvg__addShape(p);
	}
}

static void wxm_nsvg__parseEllipse(wxm_NSVGparser* p, const char** attr)
{
	float cx = 0.0f;
	float cy = 0.0f;
	float rx = 0.0f;
	float ry = 0.0f;
	int i;

	for (i = 0; attr[i]; i += 2) {
		if (!wxm_nsvg__parseAttr(p, attr[i], attr[i + 1])) {
			if (strcmp(attr[i], "cx") == 0) cx = wxm_nsvg__parseCoordinate(p, attr[i+1], wxm_nsvg__actualOrigX(p), wxm_nsvg__actualWidth(p));
			if (strcmp(attr[i], "cy") == 0) cy = wxm_nsvg__parseCoordinate(p, attr[i+1], wxm_nsvg__actualOrigY(p), wxm_nsvg__actualHeight(p));
			if (strcmp(attr[i], "rx") == 0) rx = fabsf(wxm_nsvg__parseCoordinate(p, attr[i+1], 0.0f, wxm_nsvg__actualWidth(p)));
			if (strcmp(attr[i], "ry") == 0) ry = fabsf(wxm_nsvg__parseCoordinate(p, attr[i+1], 0.0f, wxm_nsvg__actualHeight(p)));
		}
	}

	if (rx > 0.0f && ry > 0.0f) {

		wxm_nsvg__resetPath(p);

		wxm_nsvg__moveTo(p, cx+rx, cy);
		wxm_nsvg__cubicBezTo(p, cx+rx, cy+ry*wxm_NSVG_KAPPA90, cx+rx*wxm_NSVG_KAPPA90, cy+ry, cx, cy+ry);
		wxm_nsvg__cubicBezTo(p, cx-rx*wxm_NSVG_KAPPA90, cy+ry, cx-rx, cy+ry*wxm_NSVG_KAPPA90, cx-rx, cy);
		wxm_nsvg__cubicBezTo(p, cx-rx, cy-ry*wxm_NSVG_KAPPA90, cx-rx*wxm_NSVG_KAPPA90, cy-ry, cx, cy-ry);
		wxm_nsvg__cubicBezTo(p, cx+rx*wxm_NSVG_KAPPA90, cy-ry, cx+rx, cy-ry*wxm_NSVG_KAPPA90, cx+rx, cy);

		wxm_nsvg__addPath(p, 1);

		wxm_nsvg__addShape(p);
	}
}

static void wxm_nsvg__parseLine(wxm_NSVGparser* p, const char** attr)
{
	float x1 = 0.0;
	float y1 = 0.0;
	float x2 = 0.0;
	float y2 = 0.0;
	int i;

	for (i = 0; attr[i]; i += 2) {
		if (!wxm_nsvg__parseAttr(p, attr[i], attr[i + 1])) {
			if (strcmp(attr[i], "x1") == 0) x1 = wxm_nsvg__parseCoordinate(p, attr[i + 1], wxm_nsvg__actualOrigX(p), wxm_nsvg__actualWidth(p));
			if (strcmp(attr[i], "y1") == 0) y1 = wxm_nsvg__parseCoordinate(p, attr[i + 1], wxm_nsvg__actualOrigY(p), wxm_nsvg__actualHeight(p));
			if (strcmp(attr[i], "x2") == 0) x2 = wxm_nsvg__parseCoordinate(p, attr[i + 1], wxm_nsvg__actualOrigX(p), wxm_nsvg__actualWidth(p));
			if (strcmp(attr[i], "y2") == 0) y2 = wxm_nsvg__parseCoordinate(p, attr[i + 1], wxm_nsvg__actualOrigY(p), wxm_nsvg__actualHeight(p));
		}
	}

	wxm_nsvg__resetPath(p);

	wxm_nsvg__moveTo(p, x1, y1);
	wxm_nsvg__lineTo(p, x2, y2);

	wxm_nsvg__addPath(p, 0);

	wxm_nsvg__addShape(p);
}

static void wxm_nsvg__parsePoly(wxm_NSVGparser* p, const char** attr, int closeFlag)
{
	int i;
	const char* s;
	float args[2];
	int nargs, npts = 0;
	char item[64];

	wxm_nsvg__resetPath(p);

	for (i = 0; attr[i]; i += 2) {
		if (!wxm_nsvg__parseAttr(p, attr[i], attr[i + 1])) {
			if (strcmp(attr[i], "points") == 0) {
				s = attr[i + 1];
				nargs = 0;
				while (*s) {
					s = wxm_nsvg__getNextPathItem(s, item);
					args[nargs++] = (float)wxm_nsvg__atof(item);
					if (nargs >= 2) {
						if (npts == 0)
							wxm_nsvg__moveTo(p, args[0], args[1]);
						else
							wxm_nsvg__lineTo(p, args[0], args[1]);
						nargs = 0;
						npts++;
					}
				}
			}
		}
	}

	wxm_nsvg__addPath(p, (char)closeFlag);

	wxm_nsvg__addShape(p);
}

static void wxm_nsvg__parseSVG(wxm_NSVGparser* p, const char** attr)
{
	int i;
	for (i = 0; attr[i]; i += 2) {
		if (!wxm_nsvg__parseAttr(p, attr[i], attr[i + 1])) {
			if (strcmp(attr[i], "width") == 0) {
				p->image->width = wxm_nsvg__parseCoordinate(p, attr[i + 1], 0.0f, 0.0f);
			} else if (strcmp(attr[i], "height") == 0) {
				p->image->height = wxm_nsvg__parseCoordinate(p, attr[i + 1], 0.0f, 0.0f);
			} else if (strcmp(attr[i], "viewBox") == 0) {
				const char *s = attr[i + 1];
				char buf[64];
				s = wxm_nsvg__parseNumber(s, buf, 64);
				p->viewMinx = wxm_nsvg__atof(buf);
				while (*s && (wxm_nsvg__isspace(*s) || *s == '%' || *s == ',')) s++;
				if (!*s) return;
				s = wxm_nsvg__parseNumber(s, buf, 64);
				p->viewMiny = wxm_nsvg__atof(buf);
				while (*s && (wxm_nsvg__isspace(*s) || *s == '%' || *s == ',')) s++;
				if (!*s) return;
				s = wxm_nsvg__parseNumber(s, buf, 64);
				p->viewWidth = wxm_nsvg__atof(buf);
				while (*s && (wxm_nsvg__isspace(*s) || *s == '%' || *s == ',')) s++;
				if (!*s) return;
				s = wxm_nsvg__parseNumber(s, buf, 64);
				p->viewHeight = wxm_nsvg__atof(buf);
			} else if (strcmp(attr[i], "preserveAspectRatio") == 0) {
				if (strstr(attr[i + 1], "none") != 0) {
					// No uniform scaling
					p->alignType = wxm_NSVG_ALIGN_NONE;
				} else {
					// Parse X align
					if (strstr(attr[i + 1], "xMin") != 0)
						p->alignX = wxm_NSVG_ALIGN_MIN;
					else if (strstr(attr[i + 1], "xMid") != 0)
						p->alignX = wxm_NSVG_ALIGN_MID;
					else if (strstr(attr[i + 1], "xMax") != 0)
						p->alignX = wxm_NSVG_ALIGN_MAX;
					// Parse X align
					if (strstr(attr[i + 1], "yMin") != 0)
						p->alignY = wxm_NSVG_ALIGN_MIN;
					else if (strstr(attr[i + 1], "yMid") != 0)
						p->alignY = wxm_NSVG_ALIGN_MID;
					else if (strstr(attr[i + 1], "yMax") != 0)
						p->alignY = wxm_NSVG_ALIGN_MAX;
					// Parse meet/slice
					p->alignType = wxm_NSVG_ALIGN_MEET;
					if (strstr(attr[i + 1], "slice") != 0)
						p->alignType = wxm_NSVG_ALIGN_SLICE;
				}
			}
		}
	}
}

static void wxm_nsvg__parseGradient(wxm_NSVGparser* p, const char** attr, signed char type)
{
	int i;
	wxm_NSVGgradientData* grad = (wxm_NSVGgradientData*)malloc(sizeof(wxm_NSVGgradientData));
	if (grad == NULL) return;
	memset(grad, 0, sizeof(wxm_NSVGgradientData));
	grad->units = wxm_NSVG_OBJECT_SPACE;
	grad->type = type;
	if (grad->type == wxm_NSVG_PAINT_LINEAR_GRADIENT) {
		grad->linear.x1 = wxm_nsvg__coord(0.0f, wxm_NSVG_UNITS_PERCENT);
		grad->linear.y1 = wxm_nsvg__coord(0.0f, wxm_NSVG_UNITS_PERCENT);
		grad->linear.x2 = wxm_nsvg__coord(100.0f, wxm_NSVG_UNITS_PERCENT);
		grad->linear.y2 = wxm_nsvg__coord(0.0f, wxm_NSVG_UNITS_PERCENT);
	} else if (grad->type == wxm_NSVG_PAINT_RADIAL_GRADIENT) {
		grad->radial.cx = wxm_nsvg__coord(50.0f, wxm_NSVG_UNITS_PERCENT);
		grad->radial.cy = wxm_nsvg__coord(50.0f, wxm_NSVG_UNITS_PERCENT);
		grad->radial.r = wxm_nsvg__coord(50.0f, wxm_NSVG_UNITS_PERCENT);
	}

	wxm_nsvg__xformIdentity(grad->xform);

	for (i = 0; attr[i]; i += 2) {
		if (strcmp(attr[i], "id") == 0) {
			strncpy(grad->id, attr[i+1], 63);
			grad->id[63] = '\0';
		} else if (!wxm_nsvg__parseAttr(p, attr[i], attr[i + 1])) {
			if (strcmp(attr[i], "gradientUnits") == 0) {
				if (strcmp(attr[i+1], "objectBoundingBox") == 0)
					grad->units = wxm_NSVG_OBJECT_SPACE;
				else
					grad->units = wxm_NSVG_USER_SPACE;
			} else if (strcmp(attr[i], "gradientTransform") == 0) {
				wxm_nsvg__parseTransform(grad->xform, attr[i + 1]);
			} else if (strcmp(attr[i], "cx") == 0) {
				grad->radial.cx = wxm_nsvg__parseCoordinateRaw(attr[i + 1]);
			} else if (strcmp(attr[i], "cy") == 0) {
				grad->radial.cy = wxm_nsvg__parseCoordinateRaw(attr[i + 1]);
			} else if (strcmp(attr[i], "r") == 0) {
				grad->radial.r = wxm_nsvg__parseCoordinateRaw(attr[i + 1]);
			} else if (strcmp(attr[i], "fx") == 0) {
				grad->radial.fx = wxm_nsvg__parseCoordinateRaw(attr[i + 1]);
			} else if (strcmp(attr[i], "fy") == 0) {
				grad->radial.fy = wxm_nsvg__parseCoordinateRaw(attr[i + 1]);
			} else if (strcmp(attr[i], "x1") == 0) {
				grad->linear.x1 = wxm_nsvg__parseCoordinateRaw(attr[i + 1]);
			} else if (strcmp(attr[i], "y1") == 0) {
				grad->linear.y1 = wxm_nsvg__parseCoordinateRaw(attr[i + 1]);
			} else if (strcmp(attr[i], "x2") == 0) {
				grad->linear.x2 = wxm_nsvg__parseCoordinateRaw(attr[i + 1]);
			} else if (strcmp(attr[i], "y2") == 0) {
				grad->linear.y2 = wxm_nsvg__parseCoordinateRaw(attr[i + 1]);
			} else if (strcmp(attr[i], "spreadMethod") == 0) {
				if (strcmp(attr[i+1], "pad") == 0)
					grad->spread = wxm_NSVG_SPREAD_PAD;
				else if (strcmp(attr[i+1], "reflect") == 0)
					grad->spread = wxm_NSVG_SPREAD_REFLECT;
				else if (strcmp(attr[i+1], "repeat") == 0)
					grad->spread = wxm_NSVG_SPREAD_REPEAT;
			} else if (strcmp(attr[i], "xlink:href") == 0) {
				const char *href = attr[i+1];
				strncpy(grad->ref, href+1, 62);
				grad->ref[62] = '\0';
			}
		}
	}

	grad->next = p->gradients;
	p->gradients = grad;
}

static void wxm_nsvg__parseGradientStop(wxm_NSVGparser* p, const char** attr)
{
	wxm_NSVGattrib* curAttr = wxm_nsvg__getAttr(p);
	wxm_NSVGgradientData* grad;
	wxm_NSVGgradientStop* stop;
	int i, idx;

	curAttr->stopOffset = 0;
	curAttr->stopColor = 0;
	curAttr->stopOpacity = 1.0f;

	for (i = 0; attr[i]; i += 2) {
		wxm_nsvg__parseAttr(p, attr[i], attr[i + 1]);
	}

	// Add stop to the last gradient.
	grad = p->gradients;
	if (grad == NULL) return;

	grad->nstops++;
	grad->stops = (wxm_NSVGgradientStop*)realloc(grad->stops, sizeof(wxm_NSVGgradientStop)*grad->nstops);
	if (grad->stops == NULL) return;

	// Insert
	idx = grad->nstops-1;
	for (i = 0; i < grad->nstops-1; i++) {
		if (curAttr->stopOffset < grad->stops[i].offset) {
			idx = i;
			break;
		}
	}
	if (idx != grad->nstops-1) {
		for (i = grad->nstops-1; i > idx; i--)
			grad->stops[i] = grad->stops[i-1];
	}

	stop = &grad->stops[idx];
	stop->color = curAttr->stopColor;
	stop->color |= (unsigned int)(curAttr->stopOpacity*255) << 24;
	stop->offset = curAttr->stopOffset;
}

static void wxm_nsvg__startElement(void* ud, const char* el, const char** attr)
{
	wxm_NSVGparser* p = (wxm_NSVGparser*)ud;

	if (p->defsFlag) {
		// Skip everything but gradients in defs
		if (strcmp(el, "linearGradient") == 0) {
			wxm_nsvg__parseGradient(p, attr, wxm_NSVG_PAINT_LINEAR_GRADIENT);
		} else if (strcmp(el, "radialGradient") == 0) {
			wxm_nsvg__parseGradient(p, attr, wxm_NSVG_PAINT_RADIAL_GRADIENT);
		} else if (strcmp(el, "stop") == 0) {
			wxm_nsvg__parseGradientStop(p, attr);
		}
		return;
	}

	if (strcmp(el, "g") == 0) {
		wxm_nsvg__pushAttr(p);
		wxm_nsvg__parseAttribs(p, attr);
	} else if (strcmp(el, "path") == 0) {
		if (p->pathFlag)	// Do not allow nested paths.
			return;
		wxm_nsvg__pushAttr(p);
		wxm_nsvg__parsePath(p, attr);
		wxm_nsvg__popAttr(p);
	} else if (strcmp(el, "rect") == 0) {
		wxm_nsvg__pushAttr(p);
		wxm_nsvg__parseRect(p, attr);
		wxm_nsvg__popAttr(p);
	} else if (strcmp(el, "circle") == 0) {
		wxm_nsvg__pushAttr(p);
		wxm_nsvg__parseCircle(p, attr);
		wxm_nsvg__popAttr(p);
	} else if (strcmp(el, "ellipse") == 0) {
		wxm_nsvg__pushAttr(p);
		wxm_nsvg__parseEllipse(p, attr);
		wxm_nsvg__popAttr(p);
	} else if (strcmp(el, "line") == 0)  {
		wxm_nsvg__pushAttr(p);
		wxm_nsvg__parseLine(p, attr);
		wxm_nsvg__popAttr(p);
	} else if (strcmp(el, "polyline") == 0)  {
		wxm_nsvg__pushAttr(p);
		wxm_nsvg__parsePoly(p, attr, 0);
		wxm_nsvg__popAttr(p);
	} else if (strcmp(el, "polygon") == 0)  {
		wxm_nsvg__pushAttr(p);
		wxm_nsvg__parsePoly(p, attr, 1);
		wxm_nsvg__popAttr(p);
	} else  if (strcmp(el, "linearGradient") == 0) {
		wxm_nsvg__parseGradient(p, attr, wxm_NSVG_PAINT_LINEAR_GRADIENT);
	} else if (strcmp(el, "radialGradient") == 0) {
		wxm_nsvg__parseGradient(p, attr, wxm_NSVG_PAINT_RADIAL_GRADIENT);
	} else if (strcmp(el, "stop") == 0) {
		wxm_nsvg__parseGradientStop(p, attr);
	} else if (strcmp(el, "defs") == 0) {
		p->defsFlag = 1;
	} else if (strcmp(el, "svg") == 0) {
		wxm_nsvg__parseSVG(p, attr);
	}
}

static void wxm_nsvg__endElement(void* ud, const char* el)
{
	wxm_NSVGparser* p = (wxm_NSVGparser*)ud;

	if (strcmp(el, "g") == 0) {
		wxm_nsvg__popAttr(p);
	} else if (strcmp(el, "path") == 0) {
		p->pathFlag = 0;
	} else if (strcmp(el, "defs") == 0) {
		p->defsFlag = 0;
	}
}

static void wxm_nsvg__content(void* ud, const char* s)
{
	wxm_NSVG_NOTUSED(ud);
	wxm_NSVG_NOTUSED(s);
	// empty
}

static void wxm_nsvg__imageBounds(wxm_NSVGparser* p, float* bounds)
{
	wxm_NSVGshape* shape;
	shape = p->image->shapes;
	if (shape == NULL) {
		bounds[0] = bounds[1] = bounds[2] = bounds[3] = 0.0;
		return;
	}
	bounds[0] = shape->bounds[0];
	bounds[1] = shape->bounds[1];
	bounds[2] = shape->bounds[2];
	bounds[3] = shape->bounds[3];
	for (shape = shape->next; shape != NULL; shape = shape->next) {
		bounds[0] = wxm_nsvg__minf(bounds[0], shape->bounds[0]);
		bounds[1] = wxm_nsvg__minf(bounds[1], shape->bounds[1]);
		bounds[2] = wxm_nsvg__maxf(bounds[2], shape->bounds[2]);
		bounds[3] = wxm_nsvg__maxf(bounds[3], shape->bounds[3]);
	}
}

static float wxm_nsvg__viewAlign(float content, float container, int type)
{
	if (type == wxm_NSVG_ALIGN_MIN)
		return 0;
	else if (type == wxm_NSVG_ALIGN_MAX)
		return container - content;
	// mid
	return (container - content) * 0.5f;
}

static void wxm_nsvg__scaleGradient(wxm_NSVGgradient* grad, float tx, float ty, float sx, float sy)
{
	float t[6];
	wxm_nsvg__xformSetTranslation(t, tx, ty);
	wxm_nsvg__xformMultiply (grad->xform, t);

	wxm_nsvg__xformSetScale(t, sx, sy);
	wxm_nsvg__xformMultiply (grad->xform, t);
}

static void wxm_nsvg__scaleToViewbox(wxm_NSVGparser* p, const char* units)
{
	wxm_NSVGshape* shape;
	wxm_NSVGpath* path;
	float tx, ty, sx, sy, us, bounds[4], t[6], avgs;
	int i;
	float* pt;

	// Guess image size if not set completely.
	wxm_nsvg__imageBounds(p, bounds);

	if (p->viewWidth == 0) {
		if (p->image->width > 0) {
			p->viewWidth = p->image->width;
		} else {
			p->viewMinx = bounds[0];
			p->viewWidth = bounds[2] - bounds[0];
		}
	}
	if (p->viewHeight == 0) {
		if (p->image->height > 0) {
			p->viewHeight = p->image->height;
		} else {
			p->viewMiny = bounds[1];
			p->viewHeight = bounds[3] - bounds[1];
		}
	}
	if (p->image->width == 0)
		p->image->width = p->viewWidth;
	if (p->image->height == 0)
		p->image->height = p->viewHeight;

	tx = -p->viewMinx;
	ty = -p->viewMiny;
	sx = p->viewWidth > 0 ? p->image->width / p->viewWidth : 0;
	sy = p->viewHeight > 0 ? p->image->height / p->viewHeight : 0;
	// Unit scaling
	us = 1.0f / wxm_nsvg__convertToPixels(p, wxm_nsvg__coord(1.0f, wxm_nsvg__parseUnits(units)), 0.0f, 1.0f);

	// Fix aspect ratio
	if (p->alignType == wxm_NSVG_ALIGN_MEET) {
		// fit whole image into viewbox
		sx = sy = wxm_nsvg__minf(sx, sy);
		tx += wxm_nsvg__viewAlign(p->viewWidth*sx, p->image->width, p->alignX) / sx;
		ty += wxm_nsvg__viewAlign(p->viewHeight*sy, p->image->height, p->alignY) / sy;
	} else if (p->alignType == wxm_NSVG_ALIGN_SLICE) {
		// fill whole viewbox with image
		sx = sy = wxm_nsvg__maxf(sx, sy);
		tx += wxm_nsvg__viewAlign(p->viewWidth*sx, p->image->width, p->alignX) / sx;
		ty += wxm_nsvg__viewAlign(p->viewHeight*sy, p->image->height, p->alignY) / sy;
	}

	// Transform
	sx *= us;
	sy *= us;
	avgs = (sx+sy) / 2.0f;
	for (shape = p->image->shapes; shape != NULL; shape = shape->next) {
		shape->bounds[0] = (shape->bounds[0] + tx) * sx;
		shape->bounds[1] = (shape->bounds[1] + ty) * sy;
		shape->bounds[2] = (shape->bounds[2] + tx) * sx;
		shape->bounds[3] = (shape->bounds[3] + ty) * sy;
		for (path = shape->paths; path != NULL; path = path->next) {
			path->bounds[0] = (path->bounds[0] + tx) * sx;
			path->bounds[1] = (path->bounds[1] + ty) * sy;
			path->bounds[2] = (path->bounds[2] + tx) * sx;
			path->bounds[3] = (path->bounds[3] + ty) * sy;
			for (i =0; i < path->npts; i++) {
				pt = &path->pts[i*2];
				pt[0] = (pt[0] + tx) * sx;
				pt[1] = (pt[1] + ty) * sy;
			}
		}

		if (shape->fill.type == wxm_NSVG_PAINT_LINEAR_GRADIENT || shape->fill.type == wxm_NSVG_PAINT_RADIAL_GRADIENT) {
			wxm_nsvg__scaleGradient(shape->fill.gradient, tx,ty, sx,sy);
			memcpy(t, shape->fill.gradient->xform, sizeof(float)*6);
			wxm_nsvg__xformInverse(shape->fill.gradient->xform, t);
		}
		if (shape->stroke.type == wxm_NSVG_PAINT_LINEAR_GRADIENT || shape->stroke.type == wxm_NSVG_PAINT_RADIAL_GRADIENT) {
			wxm_nsvg__scaleGradient(shape->stroke.gradient, tx,ty, sx,sy);
			memcpy(t, shape->stroke.gradient->xform, sizeof(float)*6);
			wxm_nsvg__xformInverse(shape->stroke.gradient->xform, t);
		}

		shape->strokeWidth *= avgs;
		shape->strokeDashOffset *= avgs;
		for (i = 0; i < shape->strokeDashCount; i++)
			shape->strokeDashArray[i] *= avgs;
	}
}

static void wxm_nsvg__createGradients(wxm_NSVGparser* p)
{
	wxm_NSVGshape* shape;

	for (shape = p->image->shapes; shape != NULL; shape = shape->next) {
		if (shape->fill.type == wxm_NSVG_PAINT_UNDEF) {
			if (shape->fillGradient[0] != '\0') {
				float inv[6], localBounds[4];
				wxm_nsvg__xformInverse(inv, shape->xform);
				wxm_nsvg__getLocalBounds(localBounds, shape, inv);
				shape->fill.gradient = wxm_nsvg__createGradient(p, shape->fillGradient, localBounds, shape->xform, &shape->fill.type);
			}
			if (shape->fill.type == wxm_NSVG_PAINT_UNDEF) {
				shape->fill.type = wxm_NSVG_PAINT_NONE;
			}
		}
		if (shape->stroke.type == wxm_NSVG_PAINT_UNDEF) {
			if (shape->strokeGradient[0] != '\0') {
				float inv[6], localBounds[4];
				wxm_nsvg__xformInverse(inv, shape->xform);
				wxm_nsvg__getLocalBounds(localBounds, shape, inv);
				shape->stroke.gradient = wxm_nsvg__createGradient(p, shape->strokeGradient, localBounds, shape->xform, &shape->stroke.type);
			}
			if (shape->stroke.type == wxm_NSVG_PAINT_UNDEF) {
				shape->stroke.type = wxm_NSVG_PAINT_NONE;
			}
		}
	}
}

wxm_NSVGimage* wxm_nsvgParse(char* input, const char* units, float dpi)
{
	wxm_NSVGparser* p;
	wxm_NSVGimage* ret = 0;

	p = wxm_nsvg__createParser();
	if (p == NULL) {
		return NULL;
	}
	p->dpi = dpi;

	wxm_nsvg__parseXML(input, wxm_nsvg__startElement, wxm_nsvg__endElement, wxm_nsvg__content, p);

	// Create gradients after all definitions have been parsed
	wxm_nsvg__createGradients(p);

	// Scale to viewBox
	wxm_nsvg__scaleToViewbox(p, units);

	ret = p->image;
	p->image = NULL;

	wxm_nsvg__deleteParser(p);

	return ret;
}

wxm_NSVGimage* wxm_nsvgParseFromFile(const char* filename, const char* units, float dpi)
{
	FILE* fp = NULL;
	size_t size;
	char* data = NULL;
	wxm_NSVGimage* image = NULL;

	fp = fopen(filename, "rb");
	if (!fp) goto error;
	fseek(fp, 0, SEEK_END);
	size = ftell(fp);
	fseek(fp, 0, SEEK_SET);
	data = (char*)malloc(size+1);
	if (data == NULL) goto error;
	if (fread(data, 1, size, fp) != size) goto error;
	data[size] = '\0';	// Must be null terminated.
	fclose(fp);
	image = wxm_nsvgParse(data, units, dpi);
	free(data);

	return image;

error:
	if (fp) fclose(fp);
	if (data) free(data);
	if (image) wxm_nsvgDelete(image);
	return NULL;
}

wxm_NSVGpath* wxm_nsvgDuplicatePath(wxm_NSVGpath* p)
{
    wxm_NSVGpath* res = NULL;

    if (p == NULL)
        return NULL;

    res = (wxm_NSVGpath*)malloc(sizeof(wxm_NSVGpath));
    if (res == NULL) goto error;
    memset(res, 0, sizeof(wxm_NSVGpath));

    res->pts = (float*)malloc(p->npts*2*sizeof(float));
    if (res->pts == NULL) goto error;
    memcpy(res->pts, p->pts, p->npts * sizeof(float) * 2);
    res->npts = p->npts;

    memcpy(res->bounds, p->bounds, sizeof(p->bounds));

    res->closed = p->closed;

    return res;

error:
    if (res != NULL) {
        free(res->pts);
        free(res);
    }
    return NULL;
}

void wxm_nsvgDelete(wxm_NSVGimage* image)
{
	wxm_NSVGshape *snext, *shape;
	if (image == NULL) return;
	shape = image->shapes;
	while (shape != NULL) {
		snext = shape->next;
		wxm_nsvg__deletePaths(shape->paths);
		wxm_nsvg__deletePaint(&shape->fill);
		wxm_nsvg__deletePaint(&shape->stroke);
		free(shape);
		shape = snext;
	}
	free(image);
}

#endif // NANOSVG_IMPLEMENTATION

#endif // NANOSVG_H
